from fedot.core.pipelines.tuning.search_space import SearchSpace
from fedot.core.repository.operation_types_repository import OperationTypesRepository
from fedot.core.repository.tasks import TaskTypesEnum


def print_models_info(task_name):
    """ Function display models and information about it for considered task

    :param task_name: name of available task type
    """

    task = _get_task_by_name(task_name)

    repository = OperationTypesRepository(operation_type='model')

    # Filter operations
    repository_operations_list = _filter_operations_by_type(repository, task)
    for model in repository_operations_list:
        if model.id != 'custom':
            hyperparameters = SearchSpace().get_operation_parameter_range(str(model.id))
            implementation_info = model.current_strategy(task)(model.id).implementation_info
            print(f"Model name - '{model.id}'")
            print(f"Available hyperparameters to optimize with tuner - {hyperparameters}")
            print(f"Strategy implementation - {model.current_strategy(task)}")
            print(f"Model implementation - {implementation_info}\n")


def print_data_operations_info(task_name):
    """ Function display data operations and information about it for considered task

    :param task_name: name of available task type
    """

    task = _get_task_by_name(task_name)

    repository = OperationTypesRepository(operation_type='data_operation')
    # Filter operations
    repository_operations_list = _filter_operations_by_type(repository, task)
    for operation in repository_operations_list:
        hyperparameters = SearchSpace().get_operation_parameter_range(str(operation.id))
        implementation_info = operation.current_strategy(task)(operation.id).implementation_info
        print(f"Data operation name - '{operation.id}'")
        print(f"Available hyperparameters to optimize with tuner - {hyperparameters}")
        print(f"Strategy implementation - {operation.current_strategy(task)}")
        print(f"Operation implementation - {implementation_info}\n")


def _filter_operations_by_type(repository, task):
    """ Function filter operations in repository by task

    :param repository: repository with operations to filter
    :param task: task type, if None, return all models

    :return repository_operations_list: list with filtered operations
    """

    if task is None:
        # Return all operations
        repository_operations_list = repository.operations
    else:
        repository_operations_list = []
        for operation in repository.operations:
            # Every operation can have several compatible task types
            task_types = operation.task_type
            if any(task == task_type for task_type in task_types):
                repository_operations_list.append(operation)

    return repository_operations_list


def operations_for_task(task_name: str):
    """ Function filter operations by task and returns dictionary with names of
    models and data operations

    :param task_name: name of available task type

    :return dict_with_operations: dictionary with operations
        - models: appropriate models for task
        - data operations: appropriate data operations for task
    """

    task = _get_task_by_name(task_name)

    # Get models and data operations
    models_repo = OperationTypesRepository()
    data_operations_repo = OperationTypesRepository(operation_type='data_operation')

    appropriate_models = models_repo.suitable_operation(task_type=task)
    appropriate_data_operations = data_operations_repo.suitable_operation(task_type=task)

    dict_with_operations = {'model': appropriate_models,
                            'data operation': appropriate_data_operations}

    return dict_with_operations


def _get_task_by_name(task_name):
    """ Function return task by its name

    :param task_name: name of available task type
    :return task: TaskTypesEnum.<appropriate task>
    """
    task_by_name = {'regression': TaskTypesEnum.regression,
                    'classification': TaskTypesEnum.classification,
                    'clustering': TaskTypesEnum.clustering,
                    'ts_forecasting': TaskTypesEnum.ts_forecasting}
    try:
        task = task_by_name[task_name]
    except KeyError:
        raise NotImplementedError(f'Task with name "{task_name}" not available, use one of '
                                  f'the following task names: {list(task_by_name.keys())}')

    return task



import os
import random

import numpy as np

from fedot.api.main import Fedot
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.utils import fedot_project_root
from fedot.remote.infrastructure.clients.test_client import TestClient
from fedot.remote.remote_evaluator import RemoteEvaluator, RemoteTaskParams

random.seed(1)
np.random.seed(1)


# WARNING - THIS SCRIPT CAN BE EVALUATED ONLY WITH THE ACCESS TO DATAMALL SYSTEM

def clip_dataframe(df, forecast_horizon, history_size):
    # Take last n elements from dataframe to train
    dataframe_cutted = df.tail(history_size + forecast_horizon)

    return dataframe_cutted


def run_automl(data: MultiModalData, features_to_use,
               forecast_horizon: int = 10, history_size: int = 397,
               timeout: int = 1):
    """ Launch AutoML FEDOT algorithm for time series forecasting task """

    folder = os.path.join(fedot_project_root(), 'cases', 'data', 'metocean')

    connect_params = {}
    exec_params = {
        'container_input_path': os.path.join(fedot_project_root(), 'cases', 'data', 'metocean'),
        'container_output_path': os.path.join(fedot_project_root(), 'cases', 'data', 'metocean', 'remote'),
        'container_config_path': os.path.join(fedot_project_root(), 'cases', 'data', 'metocean', '.'),
        'container_image': "test",
        'timeout': 1
    }
    client = TestClient(connect_params, exec_params, output_path=os.path.join(folder, 'remote'))

    remote_task_params = RemoteTaskParams(
        mode='remote',
        dataset_name='metocean_data_train',
        max_parallel=20,
        var_names=features_to_use,
        target='sea_height'
    )

    # the following client and params can be used with DataMall system
    # connect_params = DEFAULT_CONNECT_PARAMS
    # exec_params = DEFAULT_EXEC_PARAMS
    # DataMallClient(connect_params, exec_params, output_path=os.path.join(folder, 'remote'))

    evaluator = RemoteEvaluator()
    evaluator.init(
        client=client,
        remote_task_params=remote_task_params
    )

    # Prepare parameters for algorithm launch
    composer_params = {'max_depth': 6,
                       'max_arity': 3,
                       'pop_size': 20,
                       'num_of_generations': 100,
                       'preset': 'fast_train',
                       'metric': 'rmse',
                       'cv_folds': None,
                       'validation_blocks': None}

    automl = Fedot(problem='ts_forecasting', timeout=timeout, **composer_params)

    obtained_pipeline = automl.fit(data)
    forecast = automl.forecast(data)
    return forecast, obtained_pipeline


features_to_use = ['wind_speed', 'sea_height']

data = MultiModalData.from_csv_time_series(
    file_path=f'{fedot_project_root()}/cases/data/metocean/metocean_data_train.csv',
    var_names=features_to_use,
    target_column='sea_height',
    idx_column='datetime')

forecast, obtained_pipeline = run_automl(data=data, features_to_use=['wind_speed', 'sea_height'],
                                         forecast_horizon=30,
                                         history_size=200,
                                         timeout=5)

obtained_pipeline.show()



from pathlib import Path

from fedot.api.main import Fedot
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.utils import fedot_project_root


def run_multi_modal_example(file_path: str, visualization=False) -> float:
    """
    This is an example of FEDOT use on multimodal data.
    The data is taken and adapted from Wine Reviews dataset (winemag-data_first150k):
    https://www.kaggle.com/datasets/zynicide/wine-reviews
    and contains information about wine country, region, price, etc.
    Column that contains text features is 'description'.
    Other columns contain numerical and categorical features.
    The aim is to predict wine variety, so it's a classification task.

    :param file_path: path to the file with multimodal data
    :param visualization: if True, then final pipeline will be visualised

    :return: F1 metrics of the model
    """
    task = 'classification'
    path = Path(fedot_project_root(), file_path)
    data = MultiModalData.from_csv(file_path=path, task=task, target_columns='variety', index_col=None)
    fit_data, predict_data = train_test_data_setup(data, shuffle_flag=True, split_ratio=0.7)

    automl_model = Fedot(problem=task, timeout=10)
    automl_model.fit(features=fit_data,
                     target=fit_data.target)

    prediction = automl_model.predict(predict_data)
    metrics = automl_model.get_metrics()

    if visualization:
        automl_model.current_pipeline.show()

    print(f'F1 for validation sample is {round(metrics["f1"], 3)}')

    return metrics["f1"]


if __name__ == '__main__':
    run_multi_modal_example(file_path='examples/data/multimodal_wine.csv', visualization=True)



from dataclasses import dataclass
from typing import Optional

import numpy as np

from fedot.core.repository.operation_types_repository import OperationTypesRepository
from fedot.core.repository.tasks import TaskTypesEnum


@dataclass
class SupplementaryData:
    """
    A class that stores variables for internal purposes in core during fit
    and predict. For instance, to manage dataflow properties.
    """
    # Is it data in the main branch
    is_main_target: bool = True
    # Amount of nodes, which Data visited
    data_flow_length: int = 0
    # Masked features for data
    features_mask: Optional[dict] = None
    # Last visited nodes
    previous_operations: Optional[list] = None
    # Is there a data was preprocessed or not
    was_preprocessed: bool = False
    # Collection with non-int indexes
    non_int_idx: Optional[list] = None
    # Dictionary with features and target column types
    column_types: Optional[dict] = None

    @property
    def compound_mask(self):
        """ The method allows to combine a mask with features in the form of
        one-dimensional array.
        """

        input_ids = np.array(self.features_mask.get('input_ids'), dtype=str)
        flow_lens = np.array(self.features_mask.get('flow_lens'), dtype=str)
        comp_list = np.core.defchararray.add(input_ids, flow_lens)
        return comp_list

    @property
    def flow_mask(self) -> list:
        return self.features_mask.get('flow_lens')

    def define_parents(self, unique_features_masks: np.array, task: TaskTypesEnum):
        """ Define which parent should be "Data parent" and "Model parent"
        for decompose operation

        :param unique_features_masks: unique values for mask
        :param task: task to solve
        """
        if not isinstance(self.previous_operations, list) or len(self.previous_operations) == 1:
            raise ValueError(f'Data was received from one node but at least two nodes are required')

        data_operations = OperationTypesRepository('data_operation').suitable_operation(task_type=task)

        # Which data operations was in pipeline before decompose operation
        previous_data_operation = None
        for prev_operation in self.previous_operations:
            if prev_operation in data_operations:
                previous_data_operation = prev_operation
                # Take first data operation as "Data parent"
                break

        if previous_data_operation is not None:
            # Lagged operation by default is "Data parent"
            data_ids = np.ravel(np.argwhere(np.array(self.previous_operations) == previous_data_operation))
            data_parent_id = data_ids[0]
            model_ids = np.ravel(np.argwhere(np.array(self.previous_operations) != previous_data_operation))
            model_parent_id = model_ids[0]
        else:
            model_parent_id = 0
            data_parent_id = 1

        data_parent = unique_features_masks[data_parent_id]
        model_parent = unique_features_masks[model_parent_id]

        return model_parent, data_parent



import os
from typing import Optional

import numpy as np

from h2o import h2o, H2OFrame
from h2o.automl import H2OAutoML
from tpot import TPOTClassifier, TPOTRegressor

from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.pipelines.automl_wrappers import *

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.evaluation_interfaces import EvaluationStrategy
from fedot.core.repository.tasks import TaskTypesEnum


class H2OAutoMLRegressionStrategy(EvaluationStrategy):
    __operations_by_types = {
        'h2o_regr': H2OAutoML
    }

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        self.operation_impl = self._convert_to_operation(operation_type)
        super().__init__(operation_type, params)

    def fit(self, train_data: InputData):
        ip, port = self._get_h2o_connect_config()
        h2o.init(ip=ip, port=port, name='h2o_server')

        frame = self._data_transform(train_data)

        train_frame, valid_frame = frame.split_frame(ratios=[0.85])

        # make sure that your target column is the last one
        train_columns = train_frame.columns
        if train_data.task.task_type == TaskTypesEnum.ts_forecasting:
            target_len = train_data.task.task_params.forecast_length
        else:
            target_len = 1
        target_names = train_columns[-target_len:]

        models = []

        for name in target_names:
            train_columns.remove(name)
        for name in target_names:
            model = H2OAutoML(max_models=self.params_for_fit.get("max_models"),
                              seed=self.params_for_fit.get("seed"),
                              max_runtime_secs=self.params_for_fit.get("timeout") * 60 // target_len
                              )
            model.train(x=train_columns, y=name, training_frame=train_frame)
            models.append(model.leader)

        return H2OSerializationWrapper(models)

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        res = []
        for model in trained_operation.get_estimators():
            frame = H2OFrame(predict_data.features)
            prediction = model.predict(frame)
            prediction = prediction.as_data_frame().to_numpy()
            res.append(np.ravel(prediction))
        res = np.hstack(res)
        out = self._convert_to_output(res, predict_data)
        return out

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain H2O AutoML Regression Strategy for {operation_type}')

    def _data_transform(self, data: InputData) -> H2OFrame:
        if len(data.target.shape) == 1:
            concat_data = np.concatenate((data.features, data.target.reshape(-1, 1)), 1)
        else:
            concat_data = np.concatenate((data.features, data.target.reshape(-1, data.target.shape[1])), 1)
        frame = H2OFrame(python_obj=concat_data)
        return frame

    def _get_h2o_connect_config(self):
        ip = '127.0.0.1'
        port = 8888
        return ip, port


class H2OAutoMLClassificationStrategy(EvaluationStrategy):
    __operations_by_types = {
        'h2o_class': H2OAutoML
    }

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        self.operation_impl = self._convert_to_operation(operation_type)
        self.model_class = H2OSerializationWrapper
        super().__init__(operation_type, params)

    def fit(self, train_data: InputData):
        ip, port = self._get_h2o_connect_config()

        h2o.init(ip=ip, port=port, name='h2o_server')

        frame = self._data_transform(train_data)

        train_frame, valid_frame = frame.split_frame(ratios=[0.85])

        # make sure that your target column is the last one
        train_columns = train_frame.columns
        target_name = train_columns[-1]
        train_columns.remove(target_name)
        train_frame[target_name] = train_frame[target_name].asfactor()
        model = self.operation_impl(max_models=self.params_for_fit.get("max_models"),
                                    seed=self.params_for_fit.get("seed"),
                                    max_runtime_secs=self.params_for_fit.get("timeout")*60
                                    )

        model.train(x=train_columns, y=target_name, training_frame=train_frame)
        model.leader.classes_ = np.unique(train_data.target)
        return H2OSerializationWrapper([model.leader])

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        frame = self._data_transform(predict_data)
        prediction = trained_operation.get_estimators()[0].predict(frame)
        prediction = prediction.as_data_frame().to_numpy()

        if self.output_mode == 'labels':
            prediction = prediction[:, 0]
        elif self.output_mode in ['probs', 'full_probs', 'default']:
            prediction = prediction[:, 1::]
        else:
            raise ValueError(f'Output model {self.output_mode} is not supported')
        out = self._convert_to_output(prediction, predict_data)
        return out

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain H2O AutoML Classification Strategy for {operation_type}')

    def _data_transform(self, data: InputData) -> H2OFrame:
        concat_data = np.concatenate((data.features, data.target.reshape(-1, 1)), 1)
        frame = H2OFrame(python_obj=concat_data)
        return frame

    def _get_h2o_connect_config(self):
        ip = '127.0.0.1'
        port = 8888
        return ip, port


class TPOTAutoMLRegressionStrategy(EvaluationStrategy):
    __operations_by_types = {
        'tpot_regr': TPOTRegressor
    }

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        self.operation_impl = self._convert_to_operation(operation_type)
        super().__init__(operation_type, params)

    def fit(self, train_data: InputData):
        if train_data.task.task_type == TaskTypesEnum.ts_forecasting:
            target_len = train_data.task.task_params.forecast_length
        else:
            target_len = 1
        models = []
        if len(train_data.target.shape) == 1:
            target = train_data.target.reshape(-1, 1)
        else:
            target = train_data.target

        for i in range(target.shape[1]):
            model = self.operation_impl(generations=self.params_for_fit.get('generations'),
                                        population_size=self.params_for_fit.get('population_size'),
                                        verbosity=2,
                                        random_state=42,
                                        max_time_mins=self.params_for_fit.get('timeout', 0.) // target_len
                                        )
            model.fit(train_data.features.astype(float), target.astype(float)[:, i])
            models.append(model.fitted_pipeline_)
        model = TPOTRegressionSerializationWrapper(models)
        return model

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        res = []
        features = predict_data.features.astype(float)
        for model in trained_operation.get_estimators():
            prediction = model.predict(features)
            prediction = prediction
            res.append(np.ravel(prediction))
        res = np.hstack(res)
        out = self._convert_to_output(res, predict_data)
        return out

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain H2O AutoML Regression Strategy for {operation_type}')


class TPOTAutoMLClassificationStrategy(EvaluationStrategy):
    __operations_by_types = {
        'tpot_class': TPOTClassifier
    }

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        self.operation_impl = self._convert_to_operation(operation_type)
        super().__init__(operation_type, params)

    def fit(self, train_data: InputData):
        model = self.operation_impl(generations=self.params_for_fit.get('generations'),
                                    population_size=self.params_for_fit.get('population_size'),
                                    verbosity=2,
                                    random_state=42,
                                    max_time_mins=self.params_for_fit.get('timeout', 0.)
                                    )
        model.classes_ = np.unique(train_data.target)

        model.fit(train_data.features.astype(float), train_data.target.astype(int))

        return model.fitted_pipeline_

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        n_classes = len(trained_operation.classes_)
        if self.output_mode == 'labels':
            prediction = trained_operation.predict(predict_data.features.astype(float))
        elif self.output_mode in ['probs', 'full_probs', 'default']:
            prediction = trained_operation.predict_proba(predict_data.features.astype(float))
            if n_classes < 2:
                raise NotImplementedError()
            elif n_classes == 2 and self.output_mode != 'full_probs' and len(prediction.shape) > 1:
                prediction = prediction[:, 1]
        else:
            raise ValueError(f'Output model {self.output_mode} is not supported')
        out = self._convert_to_output(prediction, predict_data)
        return out

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain H2O AutoML Classification Strategy for {operation_type}')



import warnings
from abc import abstractmethod
from typing import Optional

import numpy as np
from catboost import CatBoostClassifier, CatBoostRegressor
from lightgbm.sklearn import LGBMClassifier, LGBMRegressor
from sklearn.cluster import KMeans as SklearnKmeans
from sklearn.ensemble import (
    AdaBoostRegressor,
    ExtraTreesRegressor,
    GradientBoostingRegressor,
    RandomForestClassifier,
    RandomForestRegressor
)
from sklearn.linear_model import (
    Lasso as SklearnLassoReg,
    LinearRegression as SklearnLinReg,
    LogisticRegression as SklearnLogReg,
    Ridge as SklearnRidgeReg,
    SGDRegressor as SklearnSGD
)
from sklearn.multioutput import MultiOutputClassifier, MultiOutputRegressor
from sklearn.naive_bayes import BernoulliNB as SklearnBernoulliNB, MultinomialNB as SklearnMultinomialNB
from sklearn.neural_network import MLPClassifier
from sklearn.svm import LinearSVR as SklearnSVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from xgboost import XGBClassifier, XGBRegressor

from fedot.core.data.data import InputData, OutputData
from fedot.core.log import default_log
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.operation_types_repository import OperationTypesRepository, get_operation_type_from_id
from fedot.core.repository.tasks import TaskTypesEnum
from fedot.core.utilities.random import RandomStateHandler

warnings.filterwarnings("ignore", category=UserWarning)


class EvaluationStrategy:
    """Base class to define the evaluation strategy of Operation object:
    the certain sklearn or any other operation with fit/predict methods.

    Args:
        operation_type: ``str`` of the operation defined in operation repository
        params: hyperparameters to fit the operation with
    """

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        self.params_for_fit = params or OperationParameters()
        self.operation_id = operation_type

        self.output_mode = False

        self.log = default_log(self)

    @property
    def operation_type(self):
        return get_operation_type_from_id(self.operation_id)

    @abstractmethod
    def fit(self, train_data: InputData):
        """Main method to train the operation with the data provided

        Args:
            train_data: data used for operation training

        Returns:

        """
        raise NotImplementedError()

    @abstractmethod
    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        """Method to predict the target data for predict stage.

        Args:
            trained_operation: trained operation object
            predict_data: data to predict

        Returns:
            passed data with new predicted target
        """
        raise NotImplementedError()

    def predict_for_fit(self, trained_operation, predict_data: InputData) -> OutputData:
        """Method to predict the target data for fit stage.
        Allows to implement predict method different from main predict method
        if another behaviour for fit graph stage is needed.

        Args:
            trained_operation: trained operation object
            predict_data: data to predict
        Returns:
            passed data with new predicted target
        """
        return self.predict(trained_operation, predict_data)

    @abstractmethod
    def _convert_to_operation(self, operation_type: str):
        raise NotImplementedError()

    @property
    def implementation_info(self) -> str:
        return str(self._convert_to_operation(self.operation_type))

    @staticmethod
    def _convert_to_output(prediction, predict_data: InputData,
                           output_data_type: DataTypesEnum = DataTypesEnum.table) -> OutputData:
        """Method convert prediction into :obj:`OutputData` if it is not this type yet

        Args:
            prediction: output from model implementation
            predict_data: :obj:`InputData` used for prediction
            output_data_type: :obj:`DataTypesEnum` for output

        Returns: prediction as :obj:`OutputData`
        """

        if type(prediction) is not OutputData:
            # Wrap prediction as OutputData
            converted = OutputData(idx=predict_data.idx,
                                   features=predict_data.features,
                                   predict=prediction,
                                   task=predict_data.task,
                                   target=predict_data.target,
                                   data_type=output_data_type,
                                   supplementary_data=predict_data.supplementary_data)
        else:
            converted = prediction

        return converted


class SkLearnEvaluationStrategy(EvaluationStrategy):
    """This class defines the certain operation implementation for the sklearn operations
    defined in operation repository

    Args:
        operation_type: ``str`` of the operation defined in operation or
            data operation repositories

            .. details:: possible operations:

                - ``xgbreg``-> XGBRegressor
                - ``adareg``-> AdaBoostRegressor
                - ``gbr``-> GradientBoostingRegressor
                - ``dtreg``-> DecisionTreeRegressor
                - ``treg``-> ExtraTreesRegressor
                - ``rfr``-> RandomForestRegressor
                - ``linear``-> SklearnLinReg
                - ``ridge``-> SklearnRidgeReg
                - ``lasso``-> SklearnLassoReg
                - ``svr``-> SklearnSVR
                - ``sgdr``-> SklearnSGD
                - ``lgbmreg``-> LGBMRegressor
                - ``catboostreg``-> CatBoostRegressor
                - ``xgboost``-> XGBClassifier
                - ``logit``-> SklearnLogReg
                - ``bernb``-> SklearnBernoulliNB
                - ``multinb``-> SklearnMultinomialNB
                - ``dt``-> DecisionTreeClassifier
                - ``rf``-> RandomForestClassifier
                - ``mlp``-> MLPClassifier
                - ``lgbm``-> LGBMClassifier
                - ``catboost``-> CatBoostClassifier
                - ``kmeans``-> SklearnKmeans

        params: hyperparameters to fit the operation with
    """

    __operations_by_types = {
        'xgbreg': XGBRegressor,
        'adareg': AdaBoostRegressor,
        'gbr': GradientBoostingRegressor,
        'dtreg': DecisionTreeRegressor,
        'treg': ExtraTreesRegressor,
        'rfr': RandomForestRegressor,
        'linear': SklearnLinReg,
        'ridge': SklearnRidgeReg,
        'lasso': SklearnLassoReg,
        'svr': SklearnSVR,
        'sgdr': SklearnSGD,
        'lgbmreg': LGBMRegressor,
        'catboostreg': CatBoostRegressor,

        'xgboost': XGBClassifier,
        'logit': SklearnLogReg,
        'bernb': SklearnBernoulliNB,
        'multinb': SklearnMultinomialNB,
        'dt': DecisionTreeClassifier,
        'rf': RandomForestClassifier,
        'mlp': MLPClassifier,
        'lgbm': LGBMClassifier,
        'catboost': CatBoostClassifier,

        'kmeans': SklearnKmeans,
    }

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        self.operation_impl = self._convert_to_operation(operation_type)
        super().__init__(operation_type, params)

    def fit(self, train_data: InputData):
        """This method is used for operation training with the data provided

        Args:
            train_data: data used for operation training

        Returns:
            trained Sklearn operation
        """

        warnings.filterwarnings("ignore", category=RuntimeWarning)

        operation_implementation = self.operation_impl(**self.params_for_fit.to_dict())

        # If model doesn't support multi-output and current task is ts_forecasting
        current_task = train_data.task.task_type
        models_repo = OperationTypesRepository()
        non_multi_models = models_repo.suitable_operation(task_type=current_task,
                                                          tags=['non_multi'])
        is_model_not_support_multi = self.operation_type in non_multi_models

        # Multi-output task or not
        is_multi_target = is_multi_output_task(train_data)
        if is_model_not_support_multi and is_multi_target:
            # Manually wrap the regressor into multi-output model
            operation_implementation = convert_to_multivariate_model(operation_implementation,
                                                                     train_data)
        else:
            with RandomStateHandler():
                operation_implementation.fit(train_data.features, train_data.target)
        return operation_implementation

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        """This method used for prediction of the target data

        Args:
            trained_operation: operation object
            predict_data: data to predict

        Returns:
            passed data with new predicted target
        """
        raise NotImplementedError()

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain SKlearn strategy for {operation_type}')

    def _find_operation_by_impl(self, impl):
        for operation, operation_impl in self.__operations_by_types.items():
            if operation_impl == impl:
                return operation

    @property
    def implementation_info(self) -> str:
        return str(self._convert_to_operation(self.operation_type))

    def _sklearn_compatible_prediction(self, trained_operation, features):
        is_multi_output_target = isinstance(trained_operation.classes_, list)
        # Check if target is multilabel (has 2 or more columns)
        if is_multi_output_target:
            n_classes = len(trained_operation.classes_[0])
        else:
            n_classes = len(trained_operation.classes_)
        if self.output_mode == 'labels':
            prediction = trained_operation.predict(features)
        elif self.output_mode in ['probs', 'full_probs', 'default']:
            prediction = trained_operation.predict_proba(features)
            if n_classes < 2:
                raise NotImplementedError()
            elif n_classes == 2 and self.output_mode != 'full_probs':
                if is_multi_output_target:
                    prediction = np.stack([pred[:, 1] for pred in prediction]).T
                else:
                    prediction = prediction[:, 1]
        else:
            raise ValueError(f'Output model {self.output_mode} is not supported')

        return prediction


def convert_to_multivariate_model(sklearn_model, train_data: InputData):
    """The function returns an iterator for multiple target for those models for
    which such a function is not initially provided

    Args:
        sklearn_model: :obj:`Sklearn model` to train
        train_data: data used for model training
    Returns:
        wrapped :obj:`Sklearn model`
    """

    if train_data.task.task_type == TaskTypesEnum.classification:
        multiout_func = MultiOutputClassifier
    elif train_data.task.task_type in [TaskTypesEnum.regression, TaskTypesEnum.ts_forecasting]:
        multiout_func = MultiOutputRegressor
    else:
        raise ValueError(f"For task type '{train_data.task.task_type}' MultiOutput wrapper is not supported")

    # Apply MultiOutput
    sklearn_model = multiout_func(sklearn_model)
    sklearn_model.fit(train_data.features, train_data.target)
    return sklearn_model


def is_multi_output_task(train_data):
    target_shape = train_data.target.shape
    is_multi_target = len(target_shape) > 1 and target_shape[1] > 1
    return is_multi_target



import json
import os
import sys
from typing import Union

from fedot.core.data.data import InputData
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.log import default_log
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.verification import verifier_for_task
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import TaskTypesEnum
from fedot.core.utilities.random import RandomStateHandler
from fedot.remote.pipeline_run_config import PipelineRunConfig


def new_key_name(data_part_key):
    if data_part_key == 'idx':
        return 'idx'
    return f'data_source_ts/{data_part_key}'


def _load_ts_data(config):
    task = config.task
    if config.is_multi_modal:
        train_data = MultiModalData.from_csv_time_series(
            file_path=config.input_data,
            task=task, target_column=config.target,
            var_names=config.var_names)
    else:
        train_data = InputData.from_csv_time_series(
            file_path=config.input_data,
            task=task, target_column=config.target)
    return train_data


def _load_data(config):
    data_type = DataTypesEnum.table
    if config.task.task_type == TaskTypesEnum.ts_forecasting:
        train_data = _load_ts_data(config)
    else:
        train_data = InputData.from_csv(file_path=config.input_data,
                                        task=config.task, data_type=data_type)
    return train_data


def fit_pipeline(config_file: Union[str, bytes]) -> bool:
    logger = default_log(prefix='pipeline_fitting_logger')

    config = \
        PipelineRunConfig().load_from_file(config_file)

    verifier = verifier_for_task(config.task.task_type)

    pipeline = pipeline_from_json(config.pipeline_template)

    train_data = _load_data(config)

    # subset data using indices
    if config.train_data_idx not in [None, []]:
        train_data = train_data.subset_indices(config.train_data_idx)

    if not verifier(pipeline):
        logger.error('Pipeline not valid')
        return False

    try:
        RandomStateHandler.MODEL_FITTING_SEED = 0
        pipeline.fit_from_scratch(train_data)
    except Exception as ex:
        logger.error(ex)
        return False

    if config.test_data_path:
        test_data = InputData.from_csv(config.test_data_path)
        pipeline.predict(test_data)

    pipeline.save(path=os.path.join(config.output_path, 'fitted_pipeline'), create_subdir=False,
                  is_datetime_in_path=False)

    return True


def pipeline_from_json(json_str: str):
    json_dict = json.loads(json_str)
    pipeline = Pipeline.from_serialized(json_dict)

    return pipeline


if __name__ == '__main__':
    config_file_name_from_argv = sys.argv[1]
    fit_pipeline(config_file_name_from_argv)



import os
from datetime import timedelta
from typing import Optional, Type, TypeVar

from fedot.core.log import default_log
from fedot.core.utilities.serializable import Serializable
from fedot.core.utils import default_fedot_data_dir


G = TypeVar('G', bound=Serializable)


class Client:
    """
    Base class for remote evaluation client. It allow fitting the pipelines in external system instead local one.
    """

    def __init__(self, connect_params: dict, exec_params: dict, output_path: Optional[str] = None):
        """
        :param connect_params: parameters for connection to remote server
        :param exec_params: params for remote task execution
        :param output_path: local path for temporary saving of downloaded pipelines
        """
        self.connect_params = connect_params
        self.exec_params = exec_params
        self.output_path = output_path if output_path else \
            os.path.join(default_fedot_data_dir(), 'remote_fit_results')
        self._logger = default_log(prefix='ClientLog')

    def create_task(self, config: dict):
        """
        Create task for execution
        :param config - configuration of pipeline fitting
        :return: id of created task
        """
        raise NotImplementedError()

    def wait_until_ready(self) -> timedelta:
        """
        Delay execution until all remote tasks are ready
        :return: waiting time
        """
        raise NotImplementedError()

    def download_result(self, execution_id: int, result_cls: Type[G]) -> G:
        """
        :param execution_id: id of remote task
        :param result_cls: result
        :return: fitted pipeline downloaded from the remote server
        """
        raise NotImplementedError()



from copy import deepcopy
from typing import Any, Optional, Dict

from fedot.core.adapter import BaseOptimizationAdapter
from fedot.core.dag.graph_utils import map_dag_nodes
from fedot.core.optimisers.graph import OptGraph, OptNode
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode, Node
from fedot.core.pipelines.pipeline import Pipeline


class PipelineAdapter(BaseOptimizationAdapter[Pipeline]):
    """Optimization adapter for Pipeline<->OptGraph translation.
    It does 2 things:
    - on restore: recreate Pipeline Nodes from information stored in OptNodes
    - on adapt: create light-weight OptGraph (without 'heavy' data like
        fitted models) that can be used for reconstructing Pipelines.
    """

    def __init__(self):
        super().__init__(base_graph_class=Pipeline)

    @staticmethod
    def _transform_to_opt_node(node: Node) -> OptNode:
        # Prepare content for nodes, leave only simple data
        operation_name = str(node.operation)
        content = {'name': operation_name,
                   'params': node.parameters,
                   'metadata': node.metadata}
        return OptNode(deepcopy(content))

    @staticmethod
    def _transform_to_pipeline_node(node: OptNode) -> Node:
        # deepcopy to avoid accidental information sharing between opt graphs & pipelines
        content = deepcopy(node.content)
        if not node.nodes_from:
            return PrimaryNode(operation_type=content['name'], content=content)
        else:
            # `nodes_from` are assigned on the step of overall graph mapping
            return SecondaryNode(operation_type=content['name'], content=content)

    def _adapt(self, adaptee: Pipeline) -> OptGraph:
        adapted_nodes = map_dag_nodes(self._transform_to_opt_node, adaptee.nodes)
        return OptGraph(adapted_nodes)

    def _restore(self, opt_graph: OptGraph, metadata: Optional[Dict[str, Any]] = None) -> Pipeline:
        restored_nodes = map_dag_nodes(self._transform_to_pipeline_node, opt_graph.nodes)
        pipeline = Pipeline(restored_nodes)

        metadata = metadata or {}
        pipeline.computation_time = metadata.get('computation_time_in_seconds')
        return pipeline



import numpy as np
from hyperopt import hp


class SearchSpace:
    """
    Class for extracting searching space

    :param custom_search_space: dictionary of dictionaries of tuples (hyperopt expression (e.g. hp.choice), *params)
     for applying custom hyperparameters search space
    :param replace_default_search_space: whether replace default dictionary (False) or append it (True)
    """

    def __init__(self,
                 custom_search_space: dict = None,
                 replace_default_search_space: bool = False):
        self.custom_search_space = custom_search_space
        self.replace_default_search_space = replace_default_search_space
        self.parameters_per_operation = self.get_parameters_dict()

    def get_parameters_dict(self):
        parameters_per_operation = {
            'kmeans': {
                'n_clusters': (hp.uniformint, [2, 7])
            },
            'adareg': {

                'learning_rate': (hp.loguniform, [np.log(1e-3), np.log(1)]),
                'loss': (hp.choice, [["linear", "square", "exponential"]])
            },
            'gbr': {

                'loss': (hp.choice, [["ls", "lad", "huber", "quantile"]]),
                'learning_rate': (hp.loguniform, [np.log(1e-3), np.log(1)]),
                'max_depth': (hp.uniformint, [1, 11]),
                'min_samples_split': (hp.uniformint, [2, 21]),
                'min_samples_leaf': (hp.uniformint, [1, 21]),
                'subsample': (hp.uniform, [0.05, 1.0]),
                'max_features': (hp.uniform, [0.05, 1.0]),
                'alpha': (hp.uniform, [0.75, 0.99])
            },
            'logit': {
                'C': (hp.uniform, [1e-2, 10.0])
            },
            'rf': {
                'criterion': (hp.choice, [["gini", "entropy"]]),
                'max_features': (hp.uniform, [0.05, 1.0]),
                'min_samples_split': (hp.uniformint, [2, 10]),
                'min_samples_leaf': (hp.uniformint, [1, 15]),
                'bootstrap': (hp.choice, [[True, False]])
            },
            'lasso': {
                'alpha': (hp.uniform, [0.01, 10.0])
            },
            'ridge': {
                'alpha': (hp.uniform, [0.01, 10.0])
            },
            'rfr': {

                'max_features': (hp.uniform, [0.05, 1.0]),
                'min_samples_split': (hp.uniformint, [2, 21]),
                'min_samples_leaf': (hp.uniformint, [1, 21]),
                'bootstrap': (hp.choice, [[True, False]])
            },
            'xgbreg': {

                'max_depth': (hp.uniformint, [1, 11]),
                'learning_rate': (hp.loguniform, [np.log(1e-3), np.log(1)]),
                'subsample': (hp.uniform, [0.05, 1.0]),
                'min_child_weight': (hp.uniformint, [1, 21]),
                'objective': (hp.choice, [['reg:squarederror']])
            },
            'xgboost': {

                'max_depth': (hp.uniformint, [1, 7]),
                'learning_rate': (hp.loguniform, [np.log(1e-3), np.log(1)]),
                'subsample': (hp.uniform, [0.05, 0.99]),
                'min_child_weight': (hp.uniform, [1, 21])
            },
            'svr': {
                'loss': (hp.choice, [["epsilon_insensitive", "squared_epsilon_insensitive"]]),
                'tol': (hp.loguniform, [np.log(1e-5), np.log(1e-1)]),
                'C': (hp.uniform, [1e-4, 25.0]),
                'epsilon': (hp.uniform, [1e-4, 1.0])
            },
            'dtreg': {
                'max_depth': (hp.uniformint, [1, 11]),
                'min_samples_split': (hp.uniformint, [2, 21]),
                'min_samples_leaf': (hp.uniformint, [1, 21])
            },
            'treg': {

                'max_features': (hp.uniform, [0.05, 1.0]),
                'min_samples_split': (hp.uniformint, [2, 21]),
                'min_samples_leaf': (hp.uniformint, [1, 21]),
                'bootstrap': (hp.choice, [[True, False]])
            },
            'dt': {
                'max_depth': (hp.uniformint, [1, 11]),
                'min_samples_split': (hp.uniformint, [2, 21]),
                'min_samples_leaf': (hp.uniformint, [1, 21])
            },
            'knnreg': {
                'n_neighbors': (hp.uniformint, [1, 50]),
                'weights': (hp.choice, [["uniform", "distance"]]),
                'p': (hp.choice, [[1, 2]])
            },
            'knn': {
                'n_neighbors': (hp.uniformint, [1, 50]),
                'weights': (hp.choice, [["uniform", "distance"]]),
                'p': (hp.choice, [[1, 2]])
            },
            'arima': {
                'p': (hp.uniformint, [1, 7]),
                'd': (hp.uniformint, [0, 2]),
                'q': (hp.uniformint, [1, 5])
            },
            'stl_arima': {
                'p': (hp.uniformint, [1, 7]),
                'd': (hp.uniformint, [0, 2]),
                'q': (hp.uniformint, [1, 5]),
                'period': (hp.uniformint, [1, 365])
            },
            'ar': {
                'lag_1': (hp.uniform, [2, 200]),
                'lag_2': (hp.uniform, [2, 800])
            },
            'ets': {
                'error': (hp.choice, [['add', 'mul']]),
                'trend': (hp.choice, [[None, 'add', 'mul']]),
                'seasonal': (hp.choice, [[None, 'add', 'mul']]),
                'damped_trend': (hp.choice, [[True, False]]),
                'seasonal_periods': (hp.uniform, [1, 100])
            },
            'glm': {'nested_space': (hp.choice, [[
                {
                    'family': 'gaussian',
                    'link': hp.choice('link_gaussian', ['identity',
                                                        'inverse_power',
                                                        'log'])
                },
                {
                    'family': 'gamma',
                    'link': hp.choice('link_gamma', ['identity',
                                                     'inverse_power',
                                                     'log'])
                },
                {
                    'family': 'inverse_gaussian',
                    'link': hp.choice('link_inv_gaussian', ['identity',
                                                            'inverse_power'])
                }

            ]])},
            'clstm': {
                'window_size': (hp.uniform, [1, 200]),
                'hidden_size': (hp.uniform, [20, 200]),
                'teacher_forcing': (hp.uniform, [0, 1]),
                'learning_rate': (hp.uniform, [0.0005, 0.005]),
                'cnn1_kernel_size': (hp.uniformint, [3, 8]),
                'cnn1_output_size': (hp.choice, [[8, 16, 32, 64]]),
                'cnn2_kernel_size': (hp.uniformint, [3, 8]),
                'cnn2_output_size': (hp.choice, [[8, 16, 32, 64]]),
                'batch_size': (hp.choice, [[64, 128]]),
                'num_epochs': (hp.choice, [[10, 20, 50, 100]]),
                'optimizer': (hp.choice, [['adam', 'sgd']]),
                'loss': (hp.choice, [['mae', 'mse']])
            },
            'pca': {
                'n_components': (hp.uniform, [0.1, 0.99]),
                'svd_solver': (hp.choice, [['full']])
            },
            'kernel_pca': {
                'n_components': (hp.uniformint, [1, 20]),
                'kernel': (hp.choice, [['linear', 'poly', 'rbf', 'sigmoid', 'cosine', 'precomputed']])
            },
            'fast_ica': {
                'n_components': (hp.uniformint, [1, 20]),
                'fun': (hp.choice, [['logcosh', 'exp', 'cube']])
            },
            'ransac_lin_reg': {
                'min_samples': (hp.uniform, [0.1, 0.9]),
                'residual_threshold': (hp.loguniform, [np.log(0.1), np.log(1000)]),
                'max_trials': (hp.uniform, [50, 500]),
                'max_skips': (hp.uniform, [50, 500000])
            },
            'ransac_non_lin_reg': {
                'min_samples': (hp.uniform, [0.1, 0.9]),
                'residual_threshold': (hp.loguniform, [np.log(0.1), np.log(1000)]),
                'max_trials': (hp.uniform, [50, 500]),
                'max_skips': (hp.uniform, [50, 500000])
            },
            'isolation_forest_reg': {
                'max_samples': (hp.uniform, [0.05, 0.99]),
                'max_features': (hp.uniform, [0.05, 0.99]),
                'bootstrap': (hp.choice, [[True, False]])
            },
            'isolation_forest_class': {
                'max_samples': (hp.uniform, [0.05, 0.99]),
                'max_features': (hp.uniform, [0.05, 0.99]),
                'bootstrap': (hp.choice, [[True, False]])
            },
            'rfe_lin_reg': {
                'n_features_to_select': (hp.uniform, [0.5, 0.9]),
                'step': (hp.uniform, [0.1, 0.2])
            },
            'rfe_non_lin_reg': {
                'n_features_to_select': (hp.uniform, [0.5, 0.9]),
                'step': (hp.uniform, [0.1, 0.2])
            },
            'poly_features': {
                'degree': (hp.uniformint, [2, 5]),
                'interaction_only': (hp.choice, [[True, False]])
            },
            'polyfit': {
                'degree': (hp.uniformint, [1, 6])
            },
            'lagged': {
                'window_size': (hp.uniformint, [5, 500])
            },
            'sparse_lagged': {
                'window_size': (hp.uniformint, [5, 500]),
                'n_components': (hp.uniform, [0, 0.5]),
                'use_svd': (hp.choice, [[True, False]])
            },
            'smoothing': {
                'window_size': (hp.uniformint, [2, 20])
            },
            'gaussian_filter': {
                'sigma': (hp.uniform, [1, 5])
            },
            'diff_filter': {
                'poly_degree': (hp.uniformint, [1, 5]),
                'order': (hp.uniform, [1, 3]),
                'window_size': (hp.uniform, [3, 20])
            },
            'cut': {
                'cut_part': (hp.uniform, [0, 0.9])
            },
            'lgbm': {
                'class_weight': (hp.choice, [[None, 'balanced']]),
                'num_leaves': (hp.uniformint, [2, 256]),
                'learning_rate': (hp.loguniform, [np.log(0.01), np.log(0.2)]),
                'colsample_bytree': (hp.uniform, [0.4, 1]),
                'subsample': (hp.uniform, [0.4, 1]),
                'reg_alpha': (hp.loguniform, [np.log(1e-8), np.log(10)]),
                'reg_lambda': (hp.loguniform, [np.log(1e-8), np.log(10)])
            },
            'lgbmreg': {
                'num_leaves': (hp.uniformint, [2, 256]),
                'learning_rate': (hp.loguniform, [np.log(0.01), np.log(0.2)]),
                'colsample_bytree': (hp.uniform, [0.4, 1]),
                'subsample': (hp.uniform, [0.4, 1]),
                'reg_alpha': (hp.loguniform, [np.log(1e-8), np.log(10)]),
                'reg_lambda': (hp.loguniform, [np.log(1e-8), np.log(10)])
            },
            'catboost': {
                'max_depth': (hp.uniformint, [1, 11]),
                'learning_rate': (hp.loguniform, [np.log(0.01), np.log(0.2)]),
                'min_data_in_leaf': (hp.qloguniform, [0, 6, 1]),
                'border_count': (hp.uniformint, [2, 255]),
                'l2_leaf_reg': (hp.loguniform, [np.log(1e-8), np.log(10)]),
                'loss_function': (hp.choice, [['Logloss', 'CrossEntropy']])
            },
            'catboostreg': {
                'max_depth': (hp.uniformint, [1, 11]),
                'learning_rate': (hp.loguniform, [np.log(0.01), np.log(0.2)]),
                'min_data_in_leaf': (hp.qloguniform, [0, 6, 1]),
                'border_count': (hp.uniformint, [2, 255]),
                'l2_leaf_reg': (hp.loguniform, [np.log(1e-8), np.log(10)])
            },
            'resample': {
                'balance': (hp.choice, [['expand_minority', 'reduce_majority']]),
                'replace': (hp.choice, [[True, False]]),
                'balance_ratio': (hp.uniform, [0.3, 1])
            },
            'lda': {
                'solver': (hp.choice, [['svd', 'lsqr', 'eigen']]),
                'shrinkage': (hp.uniform, [0.1, 0.9])
            },
            'ts_naive_average': {'part_for_averaging': (hp.uniform, [0.1, 1])},
            'locf': {'part_for_repeat': (hp.uniform, [0.01, 0.5])},
            'word2vec_pretrained': {
                'model_name': (hp.choice, [['glove-twitter-25', 'glove-twitter-50',
                                            'glove-wiki-gigaword-100', 'word2vec-ruscorpora-300']])
            },
            'tfidf': {
                'ngram_range': (hp.choice, [[(1, 1), (1, 2), (1, 3)]]),
                'min_df': (hp.uniform, [0.0001, 0.1]),
                'max_df': (hp.uniform, [0.9, 0.99])
            },
        }

        if self.custom_search_space is not None:
            for operation in self.custom_search_space.keys():
                if self.replace_default_search_space:
                    parameters_per_operation[operation] = self.custom_search_space[operation]
                else:
                    for key, value in self.custom_search_space[operation].items():
                        parameters_per_operation[operation][key] = value

        return parameters_per_operation

    def get_operation_parameter_range(self, operation_name: str, parameter_name: str = None, label: str = 'default'):
        """
        Method return hyperopt object with search_space from parameters_per_operation dictionary
        If parameter name is not defined - return all available operations

        :param operation_name: name of the operation
        :param parameter_name: name of hyperparameter of particular operation
        :param label: label to assign in hyperopt pyll

        :return : dictionary with appropriate range
        """

        # Get available parameters for current operation
        operation_parameters = self.parameters_per_operation.get(operation_name)

        if operation_parameters is not None:
            # If there are not parameter_name - return list with all parameters
            if parameter_name is None:
                return list(operation_parameters)
            else:
                hyperopt_tuple = operation_parameters.get(parameter_name)
                return hyperopt_tuple[0](label, *hyperopt_tuple[1])
        else:
            return None

    def get_node_params(self, node_id, operation_name):
        """
        Method for forming dictionary with hyperparameters for considering
        operation as a part of the whole pipeline

        :param node_id: number of node in pipeline.nodes list
        :param operation_name: name of operation in the node

        :return params_dict: dictionary-like structure with labeled hyperparameters
        and their range per operation
        """

        # Get available parameters for operation
        params_list = self.get_operation_parameter_range(operation_name)

        if params_list is None:
            params_dict = None
        else:
            params_dict = {}
            for parameter_name in params_list:
                node_op_parameter_name = get_node_operation_parameter_label(node_id, operation_name, parameter_name)

                # For operation get range where search can be done
                space = self.get_operation_parameter_range(operation_name=operation_name,
                                                           parameter_name=parameter_name,
                                                           label=node_op_parameter_name)

                params_dict.update({node_op_parameter_name: space})

        return params_dict


def get_node_operation_parameter_label(node_id: int, operation_name: str, parameter_name: str) -> str:
    # Name with operation and parameter
    op_parameter_name = ''.join((operation_name, ' | ', parameter_name))

    # Name with node id || operation | parameter
    node_op_parameter_name = ''.join((str(node_id), ' || ', op_parameter_name))
    return node_op_parameter_name


def convert_params(params):
    """
    Function removes labels from dictionary with operations

    :param params: labeled parameters
    :return new_params: dictionary without labels of node_id and operation_name
    """

    new_params = {}
    for operation_parameter, value in params.items():
        # Remove right part of the parameter name
        parameter_name = operation_parameter.split(' | ')[-1]

        if value is not None:
            new_params.update({parameter_name: value})

    return new_params



from typing import Tuple, List, Dict, Any

import networkx as nx
import numpy as np
import pandas as pd

from fedot.core.dag.convert import graph_structure_as_nx_graph

DEFAULT_SOURCE_NAME = 'default'


class PipelineStructureExplorer:
    """ Class for performing pipeline structure exploration.
    The class allows you to convert pipelines into a networkx graph and considers
    all possible paths from PrimaryNode (or PrimaryNodes) to root node. If at least
    one of the paths contains an invalid sequence of operations, the search performed
    by this class will detect it
    """

    _invariant_tags = {'encoding': 'categorical-ignore',
                       'imputation': 'nans-ignore'}

    @staticmethod
    def check_structure_by_tag(pipeline: 'Pipeline', tag_to_check: str, source_name: str = DEFAULT_SOURCE_NAME):
        """
        In the pipeline structure, a node with an operation with the appropriate tag is searched for.
        In this case the operations must have priority in the pipeline - in the PrimaryNode or not far from it.

        Correct pipeline:
        operation with tag -> linear
        Incorrect pipeline:
        linear -> operation with tag

        :param pipeline: pipeline to check
        :param tag_to_check: find appropriate operation by desired tag
        (for example encoding or imputing)
        :param source_name: label of primary node for current input data. Set as 'default' if
        pipeline prepared for unimodal data
        """
        if len(pipeline.nodes) < 2:
            # Preprocessing needed for single-node pipeline
            return False

        graph, node_labels = graph_structure_as_nx_graph(pipeline)

        # Assign information for all nodes in the graph
        graph, info_df = PipelineStructureExplorer._enrich_with_information(graph, node_labels)

        primary_df = info_df[info_df['node_type'] == 'primary']
        root_df = info_df[info_df['node_type'] == 'root']
        root_id = root_df.iloc[0]['node_id']

        paths = {}
        path_id = 0
        for i, node_info in primary_df.iterrows():
            primary_id = node_info['node_id']
            node_name = node_info['node_label'].operation.operation_type
            if source_name in (node_name, DEFAULT_SOURCE_NAME):
                for path in nx.all_simple_paths(graph, source=primary_id, target=root_id):
                    # Check the path (branch) whether it has wanted operation in correct location or not
                    path_info = PipelineStructureExplorer.check_path(graph, path, tag_to_check)
                    paths[path_id] = path_info
                    path_id += 1

        correct_branches = (branch['correctness'] for branch in paths.values())
        # 'False' means that least one branch in the graph cannot process desired type of data
        return all(correct_branches)

    @staticmethod
    def check_path(graph: nx.DiGraph, path: list, tag_to_check: str) -> Dict[str, Any]:
        """
        Checking the path for operations take right places in the pipeline.

        :param graph: graph for checking paths
        :param path: path in the graph from PrimaryNode to root
        :param tag_to_check: find appropriate operation by desired tag
        """
        operation_path, is_appropriate_operation, is_independent_operation = \
            PipelineStructureExplorer._calculate_binary_paths(graph, path, tag_to_check)

        # Define is branch correct or not
        is_branch_correct = PipelineStructureExplorer.is_current_branch_correct(is_appropriate_operation,
                                                                                is_independent_operation,
                                                                                len(path))

        path_info = {'correctness': is_branch_correct,
                     'path': operation_path}

        return path_info

    @staticmethod
    def _enrich_with_information(graph: nx.DiGraph, node_labels: dict) -> Tuple[nx.DiGraph, pd.DataFrame]:
        """
        Set additional information (operation name and node type) to nodes as attributes.
        There is also preparing pandas DataFrame with such information

        :param node_labels: dictionary with ids and operation names
        """
        # Set names to nodes and additional info
        number_of_out_edges = graph.degree._nodes
        number_of_in_edges = graph.in_edges._adjdict

        info_df = []
        for node_id, node_label in node_labels.items():
            parent_numbers = len(number_of_in_edges[node_id])
            child_numbers = len(number_of_out_edges[node_id])

            if parent_numbers == 0:
                node_type = 'primary'
            else:
                if child_numbers == 0:
                    # It is final node in the pipeline
                    node_type = 'root'
                else:
                    node_type = 'secondary'

            attrs = {node_id: {'operation': node_label, 'type': node_type}}
            nx.set_node_attributes(graph, attrs)

            info_df.append([node_id, node_type, node_label, parent_numbers, child_numbers])

        info_df = pd.DataFrame(info_df, columns=['node_id', 'node_type', 'node_label',
                                                 'parent_number', 'child_number'])
        return graph, info_df

    @staticmethod
    def _calculate_binary_paths(graph: nx.Graph, path: list, tag_to_check: str) \
            -> Tuple[List[str], List[bool], List[bool]]:
        """
        Calculate binary masks for considering path in the graph.
        For example, branch
        UID 584fef54 -> UID 568984edr45 -> UID 4566895ef13

        Can be decomposed as follows (if tag_to_check is encoding):

            * operations_path:
            imputation -> encoding -> ridge
            Listing the names of the operations in the nodes

            * is_appropriate_operation:
            False -> True -> False
            Whether there is an operation with a matching tag in the node or not

            * is_independent_operation
            True -> False -> False
            Are there any operations in the branch that can process data without
            errors, even if they have not been previously processed by encoding,
            e.g.
        """
        ignore_tag = PipelineStructureExplorer._invariant_tags.get(tag_to_check)

        # List with names of operations in the branch
        operations_path = []
        # Is the operation contain desired tag or not
        is_appropriate_operation = []
        # Can operation ignore errors when processing data without applying "desired operation"
        # For example: "desired operation" is imputing. If considering operation can process
        # data with nans in can_be_ignored list "True" will be added
        is_independent_operation = []
        for node_id in path:
            current_node = graph.nodes.get(node_id)

            node_tags = current_node['operation'].tags
            if tag_to_check in node_tags:
                is_appropriate_operation.append(True)
            else:
                # Operation in the node is not wanted one
                is_appropriate_operation.append(False)

                if ignore_tag is not None:
                    if ignore_tag in node_tags:
                        is_independent_operation.append(True)
                    else:
                        is_independent_operation.append(False)

            operations_path.append(current_node['operation'])

        return operations_path, is_appropriate_operation, is_independent_operation

    @staticmethod
    def is_current_branch_correct(is_appropriate_operation,
                                  is_independent_operation,
                                  path_len):
        """
        Based on predefined rules, the branch is checked for correctness.
        1) is_appropriate_operation
        2) is_independent_operation
        For example correct branch is:
            1) False -> True -> False
            2) True -> False -> False
            Data can be processed without errors due to wanted operation applied
            after independent operation
            True -> True -> after operation output everything is ok

        Incorrect branch is:
            1) False -> False -> True
            2) True -> False -> False
            Second operation is not wanted one and at the same time cannot
            process data without any transformation. So it will have an error
            True -> X False X -> True

        :param is_appropriate_operation: list with bool values is wanted operation placed in the node or not
        :param is_independent_operation: list with bool values is independent operation placed in the node or not
        :param path_len: length of list with graph nodes
        """

        # Check if operations were in the path
        is_appropriate_operation = np.array(is_appropriate_operation)
        # Number of True in list
        number_operations = is_appropriate_operation.sum()

        # True > 0 - so find ids with appropriate operations
        operation_ids = np.ravel(np.argwhere(is_appropriate_operation > 0))

        # Find independent operations in the path
        if len(is_independent_operation) > 0:
            is_independent_operation = np.array(is_independent_operation)
        else:
            is_independent_operation = np.array([False] * path_len)

        # Check independent operation presence in path
        number_independent_operations = np.sum(is_independent_operation)

        # By default it is incorrect
        is_branch_correct = False
        if number_operations == 0 and number_independent_operations == 0:
            # Path contain non independent operations and have no good ones
            is_branch_correct = False

        elif number_operations > 0 and number_independent_operations == 0:
            # Path contain good operations but have no independent
            if operation_ids[0] == 0:
                # Wanted operation in the first place
                is_branch_correct = True
            else:
                is_branch_correct = False

        elif number_operations == 0 and number_independent_operations > 0:
            # Have no wanted operations but have independent operations
            if number_independent_operations == path_len:
                # All operations are independent
                is_branch_correct = True
            else:
                is_branch_correct = False

        elif number_operations > 0 and number_independent_operations > 0:
            # Have both independent operation(s) and wanted operations
            path_before_good_operation = is_independent_operation[:operation_ids[0]]
            if path_before_good_operation.sum() == len(path_before_good_operation):
                # All operations are independent before wanted operation applying
                is_branch_correct = True
            else:
                is_branch_correct = False

        return is_branch_correct



import random

import numpy as np

from fedot.utilities.requirements_notificator import warn_requirement

try:
    import tensorflow as tf
except ModuleNotFoundError:
    warn_requirement('tensorflow')

from sklearn.metrics import roc_auc_score as roc_auc

from examples.simple.classification.classification_pipelines import cnn_composite_pipeline
from fedot.core.data.data import InputData, OutputData
from fedot.core.repository.tasks import Task, TaskTypesEnum

random.seed(1)
np.random.seed(1)


def calculate_validation_metric(predicted: OutputData, dataset_to_validate: InputData) -> float:
    # the quality assessment for the simulation results
    roc_auc_value = roc_auc(y_true=dataset_to_validate.target,
                            y_score=predicted.predict,
                            multi_class="ovo")
    return roc_auc_value


def run_image_classification_problem(train_dataset: tuple,
                                     test_dataset: tuple,
                                     composite_flag: bool = True):
    task = Task(TaskTypesEnum.classification)

    x_train, y_train = train_dataset[0], train_dataset[1]
    x_test, y_test = test_dataset[0], test_dataset[1]

    dataset_to_train = InputData.from_image(images=x_train,
                                            labels=y_train,
                                            task=task)
    dataset_to_validate = InputData.from_image(images=x_test,
                                               labels=y_test,
                                               task=task)

    pipeline = cnn_composite_pipeline(composite_flag)
    pipeline.fit(input_data=dataset_to_train)
    predictions = pipeline.predict(dataset_to_validate)
    roc_auc_on_valid = calculate_validation_metric(predictions,
                                                   dataset_to_validate)
    return roc_auc_on_valid, dataset_to_train, dataset_to_validate


if __name__ == '__main__':
    training_set, testing_set = tf.keras.datasets.mnist.load_data(path='mnist.npz')
    roc_auc_on_valid, dataset_to_train, dataset_to_validate = run_image_classification_problem(
        train_dataset=training_set,
        test_dataset=testing_set)



from copy import copy
from typing import Optional, Union

import numpy as np
import pandas as pd
from scipy.ndimage import gaussian_filter
from sklearn.decomposition import TruncatedSVD

from fedot.core.data.data import InputData, OutputData
from fedot.core.log import LoggerAdapter, default_log
from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import (
    DataOperationImplementation
)
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum


class LaggedImplementation(DataOperationImplementation):
    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)

        self.window_size_minimum = None
        self.sparse_transform = False
        self.use_svd = False
        self.features_columns = None

        # Define logger object
        self.log = default_log(self)

    @property
    def window_size(self) -> Optional[int]:
        window_size = self.params.get('window_size')
        if window_size:
            window_size = round(window_size)
        return window_size

    @property
    def n_components(self) -> Optional[int]:
        return self.params.get('n_components')

    def fit(self, input_data):
        """ Class doesn't support fit operation

        Args:
            input_data: data with features, target and ids to process
        """

        pass

    def transform(self, input_data: InputData) -> OutputData:
        """ Method for transformation of time series to lagged form for predict stage

        Args:
            input_data: data with features, target and ids to process

        Returns:
            output data with transformed features table
        """

        new_input_data = copy(input_data)
        forecast_length = new_input_data.task.task_params.forecast_length

        # Correct window size parameter
        self._check_and_correct_window_size(new_input_data.features, forecast_length)

        self._apply_transformation_for_predict(new_input_data)

        output_data = self._convert_to_output(new_input_data,
                                              self.features_columns,
                                              data_type=DataTypesEnum.table)
        self._update_column_types(output_data)
        return output_data

    def transform_for_fit(self, input_data: InputData) -> OutputData:
        """Method for transformation of time series to lagged form for fit stage

        Args:
            input_data: data with features, target and ids to process

        Returns:
            output data with transformed features table
        """
        new_input_data = copy(input_data)
        forecast_length = new_input_data.task.task_params.forecast_length

        # Correct window size parameter
        self._check_and_correct_window_size(new_input_data.features, forecast_length)

        target = np.array(new_input_data.target)
        features = np.array(new_input_data.features)
        old_idx = new_input_data.idx

        new_target, new_idx = self._apply_transformation_for_fit(new_input_data, features,
                                                                 target, forecast_length, old_idx)

        # Update target for Input Data
        new_input_data.target = new_target
        new_input_data.idx = new_idx
        output_data = self._convert_to_output(new_input_data,
                                              self.features_columns,
                                              data_type=DataTypesEnum.table)
        self._update_column_types(output_data)
        return output_data

    def _check_and_correct_window_size(self, time_series: np.array, forecast_length: int):
        """ Method check if the length of the time series is not enough for
            lagged transformation - clip it

            Args:
                time_series: time series for transformation
                forecast_length: forecast length

            Returns:

            """
        prefix = "Warning: window size of lagged transformation was changed"

        # Maximum threshold
        removing_len = self.window_size + forecast_length
        if removing_len > len(time_series):
            previous_size = self.window_size
            # At least 10 objects we need for training, so minus 10
            window_size = len(time_series) - forecast_length - 10
            self.params.update(window_size=window_size)
            self.log.info(f"{prefix} from {previous_size} to {self.window_size}.")

        # Minimum threshold
        if self.window_size < self.window_size_minimum:
            previous_size = self.window_size
            self.params.update(window_size=self.window_size_minimum)
            self.log.info(f"{prefix} from {previous_size} to {self.window_size}")

    def _update_column_types(self, output_data: OutputData):
        """Update column types after lagged transformation. All features becomes ``float``
        """

        features_n_rows, features_n_cols = output_data.predict.shape
        features_column_types = [str(float)] * features_n_cols
        column_types = {'features': features_column_types}

        if output_data.target is not None and len(output_data.target.shape) > 1:
            target_n_rows, target_n_cols = output_data.target.shape
            column_types.update({'target': [str(float)] * target_n_cols})
        output_data.supplementary_data.column_types = column_types

    def _apply_transformation_for_fit(self, input_data: InputData, features: np.array, target: np.array,
                                      forecast_length: int, old_idx: np.array):
        """Apply lagged transformation on each time series in the current dataset
        """

        # Shape of the time series
        if len(features.shape) > 1:
            # Multivariate time series
            n_elements, n_time_series = features.shape
        else:
            n_time_series = 1
        all_transformed_features = None
        for current_ts_id in range(n_time_series):
            # For each time series in the array
            if n_time_series == 1:
                current_ts = features
            else:
                current_ts = np.ravel(features[:, current_ts_id])

            # Prepare features for training
            new_idx, transformed_cols = ts_to_table(idx=old_idx,
                                                    time_series=current_ts,
                                                    window_size=self.window_size,
                                                    is_lag=True)

            # Sparsing matrix of lagged features
            if self.sparse_transform:
                transformed_cols = _sparse_matrix(self.log,
                                                  transformed_cols,
                                                  self.n_components,
                                                  self.use_svd)
            # Transform target
            current_target = self._current_target_for_each_ts(current_ts_id, target)
            new_idx, transformed_cols, new_target = prepare_target(all_idx=input_data.idx,
                                                                   idx=new_idx,
                                                                   features_columns=transformed_cols,
                                                                   target=current_target,
                                                                   forecast_length=forecast_length)
            if current_ts_id == 0:
                # Init full lagged table
                all_transformed_features = transformed_cols
                all_transformed_target = new_target
                all_transformed_idx = np.array(new_idx)
            else:
                all_transformed_features, all_transformed_target, all_transformed_idx = self.stack_by_type_fit(
                    input_data, all_transformed_features,
                    all_transformed_target, all_transformed_idx,
                    transformed_cols, new_target, new_idx)

        input_data.features = all_transformed_features
        self.features_columns = all_transformed_features
        return all_transformed_target, all_transformed_idx

    def stack_by_type_fit(self, input_data, all_features, all_target, all_idx, features, target, idx):
        """Apply stack function for multi_ts and multivariable ts types on fit step
        """

        functions_by_type = {
            DataTypesEnum.multi_ts: self._stack_multi_ts,
            DataTypesEnum.ts: self._stack_multi_variable
        }
        stack_function = functions_by_type.get(input_data.data_type)
        return stack_function(all_features, all_target, all_idx, features, target, idx)

    def _stack_multi_variable(self, all_features: np.array,
                              all_target: np.array,
                              all_idx: np.array,
                              features: np.array,
                              target: np.array,
                              idx: Union[list, np.array]):
        """Horizontally stack tables as multiple variables extends features for training

        Args:
            all_features: ``array`` with all features for adding new
            all_target:  ``array`` with all target (does not change)
            all_idx: ``array`` with all indices (does not change)
            features: ``array`` with new features for adding
            target: ``array`` with new target for adding
            idx: ``array`` with new idx for adding

        Returns:
            table
        """

        all_features = np.hstack((all_features, features))
        return all_features, all_target, all_idx

    def _stack_multi_ts(self, all_features: np.array,
                        all_target: np.array,
                        all_idx: np.array,
                        features: np.array,
                        target: np.array,
                        idx: Union[list, np.array]):
        """Vertically stack tables as multi_ts data extends training set as combination of train and target

        Args:
            all_features: ``array`` with all features for adding new
            all_target:  ``array`` with all target
            all_idx: ``array`` with all indices
            features: ``array`` with new features for adding
            target: ``array`` with new target for adding
            idx: ``array`` with new idx for adding

        Returns:
            table
        """

        all_features = np.vstack((all_features, features))
        all_target = np.vstack((all_target, target))
        all_idx = np.hstack((all_idx, np.array(idx)))
        return all_features, all_target, all_idx

    def _current_target_for_each_ts(self, current_ts_id, target):
        """Returns target for each time-series
        """

        if len(target.shape) > 1:
            # if multi_ts case
            if current_ts_id >= target.shape[1]:
                while current_ts_id >= target.shape[1]:
                    current_ts_id = current_ts_id - target.shape[1]
            return target[:, current_ts_id]
        else:
            # if multivariable case
            return target

    def _apply_transformation_for_predict(self, input_data: InputData):
        """Apply lagged transformation for every column (time series) in the dataset
        """
        old_idx = copy(input_data.idx)

        if len(input_data.features.shape) > 1:
            # Multivariate time series
            n_elements, n_time_series = input_data.features.shape
        else:
            n_time_series = 1

        all_transformed_features = None
        for current_ts_id in range(n_time_series):
            # For each time series
            if n_time_series == 1:
                current_ts = input_data.features
            else:
                current_ts = np.ravel(input_data.features[:, current_ts_id])

            if self.sparse_transform:
                self.log.debug('Sparse lagged transformation applied. If new data were used. Call fit method')
                transformed_cols = self._update_features_for_sparse(current_ts, old_idx)
                # Take last row in the lagged table and reshape into array with 1 row and n columns
                current_ts = transformed_cols[-1].reshape(1, -1)

            # Take last window_size elements for current ts
            last_part_of_ts = current_ts[-self.window_size:].reshape(1, -1)
            if current_ts_id == 0:
                all_transformed_features = last_part_of_ts
            else:
                all_transformed_features = self.stack_by_type_predict(input_data,
                                                                      all_transformed_features,
                                                                      last_part_of_ts)

        if input_data.data_type == DataTypesEnum.multi_ts:
            all_transformed_features = np.expand_dims(all_transformed_features[0], axis=0)
        self.features_columns = all_transformed_features
        return all_transformed_features

    def stack_by_type_predict(self, input_data, all_features, part_to_add):
        """Apply stack function for multi_ts and multivariable ts types on predict step
        """

        if input_data.data_type == DataTypesEnum.multi_ts:
            # for mutli_ts
            all_features = np.vstack((all_features, part_to_add))
        if input_data.data_type == DataTypesEnum.ts:
            # for multivariable
            all_features = np.hstack((all_features, part_to_add))
        return all_features

    def _update_features_for_sparse(self, time_series: np.array, idx: np.array):
        """Make sparse matrix which will be used during forecasting
        """

        # Prepare features for training
        new_idx, transformed_cols = ts_to_table(idx=idx,
                                                time_series=time_series,
                                                window_size=self.window_size,
                                                is_lag=True)
        # Sparsing
        transformed_cols = _sparse_matrix(self.log,
                                          transformed_cols,
                                          self.n_components,
                                          self.use_svd)
        return transformed_cols


class SparseLaggedTransformationImplementation(LaggedImplementation):
    """Implementation of sparse lagged transformation for time series forecasting
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.sparse_transform = True
        self.window_size_minimum = 6


class LaggedTransformationImplementation(LaggedImplementation):
    """Implementation of lagged transformation for time series forecasting
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.window_size_minimum = 2


class TsSmoothingImplementation(DataOperationImplementation):

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)

    @property
    def window_size(self) -> int:
        return round(self.params.setdefault('window_size', 10))

    def fit(self, input_data: InputData):
        """Class doesn't support fit operation

        Args:
            input_data: data with features, target and ids to process
        """

        pass

    def transform(self, input_data: InputData) -> OutputData:
        """Method for smoothing time series

        Args:
            input_data: data with features, target and ids to process

        Returns:
            output data with smoothed time series
        """

        source_ts = input_data.features
        if input_data.data_type == DataTypesEnum.multi_ts:
            full_smoothed_ts = []
            for ts_n in range(source_ts.shape[1]):
                ts = pd.Series(source_ts[:, ts_n])
                smoothed_ts = self._apply_smoothing_to_series(ts)
                full_smoothed_ts.append(smoothed_ts)
            output_data = self._convert_to_output(input_data,
                                                  np.array(full_smoothed_ts).T,
                                                  data_type=input_data.data_type)
        else:
            source_ts = pd.Series(input_data.features)
            smoothed_ts = np.ravel(self._apply_smoothing_to_series(source_ts))
            output_data = self._convert_to_output(input_data,
                                                  smoothed_ts,
                                                  data_type=input_data.data_type)

        return output_data

    def _apply_smoothing_to_series(self, ts):
        smoothed_ts = ts.rolling(window=self.window_size).mean()
        smoothed_ts = np.array(smoothed_ts)

        # Filling first nans with source values
        smoothed_ts[:self.window_size] = ts[:self.window_size]
        return smoothed_ts


class ExogDataTransformationImplementation(DataOperationImplementation):

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)

    def fit(self, input_data: InputData):
        """ Class doesn't support fit operation

        Args:
            input_data: data with features, target and ids to process
        """

        pass

    def transform(self, input_data: InputData) -> OutputData:
        """ Method for representing time series as column

        Args:
            input_data: data with features, target and ids to process

        Returns:
            output data with features as columns
        """
        copied_data = copy(input_data)
        parameters = copied_data.task.task_params
        forecast_length = parameters.forecast_length

        features_columns = np.array(copied_data.features)[-forecast_length:]
        copied_data.idx = copied_data.idx[-forecast_length:]
        features_columns = features_columns.reshape(1, -1)

        output_data = self._convert_to_output(copied_data,
                                              features_columns,
                                              data_type=DataTypesEnum.table)

        return output_data

    def transform_for_fit(self, input_data: InputData) -> OutputData:
        """Method for representing time series as column for fit stage

        Args:
            input_data: data with features, target and ids to process
        Returns:
            output data with features as columns
        """

        copied_data = copy(input_data)
        parameters = copied_data.task.task_params
        old_idx = copied_data.idx
        forecast_length = parameters.forecast_length

        # Transform features in "target-like way"
        _, _, features_columns = prepare_target(all_idx=input_data.idx,
                                                idx=old_idx,
                                                features_columns=copied_data.features,
                                                target=copied_data.features,
                                                forecast_length=forecast_length)

        # Transform target
        new_idx, _, new_target = prepare_target(all_idx=input_data.idx,
                                                idx=old_idx,
                                                features_columns=copied_data.features,
                                                target=copied_data.target,
                                                forecast_length=forecast_length)
        copied_data.target = new_target
        copied_data.idx = new_idx

        output_data = self._convert_to_output(copied_data,
                                              features_columns,
                                              data_type=DataTypesEnum.table)

        return output_data


class GaussianFilterImplementation(DataOperationImplementation):
    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)

    def fit(self, input_data: InputData):
        """ Class doesn't support fit operation

        Args:
            input_data: data with features, target and ids to process
        """

        pass

    def transform(self, input_data: InputData) -> OutputData:
        """ Method for smoothing time series for predict stage

        Args:
            input_data: data with features, target and ids to process

        Returns:
            output data with smoothed time series
        """

        source_ts = np.array(input_data.features)

        # Apply smoothing operation
        sigma = round(self.params.setdefault('sigma', 1))
        smoothed_ts = gaussian_filter(source_ts, sigma=sigma)
        smoothed_ts = np.array(smoothed_ts)

        if input_data.data_type != DataTypesEnum.multi_ts:
            smoothed_ts = np.ravel(smoothed_ts)
        output_data = self._convert_to_output(input_data,
                                              smoothed_ts,
                                              data_type=input_data.data_type)

        return output_data


class NumericalDerivativeFilterImplementation(DataOperationImplementation):
    def __init__(self, params: OperationParameters):
        super().__init__(params)
        self.max_poly_degree = 5
        self.default_poly_degree = 2
        self.default_order = 1

        self.log = default_log(self)

        self._correct_params()

    @property
    def poly_degree(self) -> int:
        return int(self.params.get('poly_degree'))

    @property
    def order(self) -> int:
        return int(self.params.get('order'))

    @property
    def window_size(self) -> int:
        return int(self.params.get('window_size'))

    def fit(self, input_data: InputData):
        """ Class doesn't support fit operation

        Args:
            input_data: data with features, target and ids to process
        """

        pass

    def transform(self, input_data: InputData) -> OutputData:
        """ Method for finding numerical derivative of time series for predict stage

        Args:
            input_data: data with features, target and ids to process

        Returns:
            output data with smoothed time series
        """

        source_ts = np.array(input_data.features)
        # Apply differential operation
        if input_data.data_type == DataTypesEnum.multi_ts:
            full_differential_ts = []
            for ts_n in range(source_ts.shape[1]):
                ts = source_ts[:, ts_n]
                differential_ts = self._differential_filter(ts)
                full_differential_ts.append(differential_ts)
            output_data = self._convert_to_output(input_data,
                                                  np.array(full_differential_ts).T,
                                                  data_type=input_data.data_type)
        else:
            differential_ts = np.ravel(self._differential_filter(source_ts))
            output_data = self._convert_to_output(input_data,
                                                  differential_ts,
                                                  data_type=input_data.data_type)

        return output_data

    def _differential_filter(self, ts):
        """:obj:`NumericalDerivative` filter
        """

        if self.window_size > ts.shape[0]:
            self.log.info(f'NumericalDerivativeFilter: invalid parameter window_size ({self.window_size}) changed to '
                          f'{self.poly_degree + 1}')
            self.params.update(window_size=self.poly_degree + 1)
        x = np.arange(ts.shape[0])

        ts_len = x.shape[0]
        der_f = np.zeros(ts_len)

        # Take the differentials in the center of the domain
        for center_window in range(self.window_size, ts_len - self.window_size):
            points = np.arange(center_window - self.window_size, center_window + self.window_size)
            # Fit to a Chebyshev polynomial
            # this is the same as any polynomial since we're on a fixed grid but it's better conditioned :)
            poly = np.polynomial.chebyshev.Chebyshev.fit(x[points], ts[points], self.poly_degree,
                                                         window=[np.min(points), np.max(points)])
            der_f[center_window] = poly.deriv(m=self.order)(x[center_window])

        supp_1 = ts[0:self.window_size]
        coordsupp_1 = x[0:self.window_size]
        supp_2 = ts[-self.window_size:]
        coordsupp_2 = x[-self.window_size:]
        poly = np.polynomial.chebyshev.Chebyshev.fit(coordsupp_1, supp_1, self.window_size - 1)
        der_f[0:self.window_size] = poly.deriv(m=self.order)(coordsupp_1)
        poly = np.polynomial.chebyshev.Chebyshev.fit(coordsupp_2, supp_2, self.window_size - 1)
        der_f[-self.window_size:] = poly.deriv(m=self.order)(coordsupp_2)
        for _ in range(self.order):
            supp_1 = np.gradient(supp_1, coordsupp_1, edge_order=2)
            supp_2 = np.gradient(supp_2, coordsupp_2, edge_order=2)
        der_f[0:self.window_size] = supp_1
        der_f[-self.window_size:] = supp_2
        return np.transpose(der_f)

    def _correct_params(self):
        if self.poly_degree > 5:
            self.log.info(f'NumericalDerivativeFilter: invalid parameter poly_degree ({self.poly_degree}) '
                          f'changed to {self.max_poly_degree}')
            self.params.update(poly_degree=self.max_poly_degree)
        if self.order < 1:
            self.log.info(f'NumericalDerivativeFilter: invalid parameter order ({self.order}) '
                          f'changed to {self.default_order}')
            self.params.update(degree=self.default_order)
        if self.order >= self.poly_degree:
            self.log.info(f'NumericalDerivativeFilter: invalid parameter poly_degree ({self.poly_degree}) '
                          f'changed to {self.order + 1}')
            self.params.update(poly_degree=self.order + 1)
        if self.window_size < self.poly_degree:
            self.log.info(f'NumericalDerivativeFilter: invalid parameter window_size ({self.window_size}) changed to '
                          f'{self.poly_degree + 1}')
            self.params.update(window_size=self.poly_degree + 1)


class CutImplementation(DataOperationImplementation):
    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self._correct_cut_part()
        # Define logger object
        self.log = default_log(self)

    @property
    def cut_part(self):
        return self.params.get('cut_part')

    def _correct_cut_part(self):
        if not 0 < self.cut_part <= 0.9:
            # Default parameter
            self.log.info(f"Change invalid parameter cut_part ({self.cut_part}) on default value (0.5)")
            self.params.update(cut_part=0.5)

    def fit(self, input_data: InputData):
        """Class doesn't support fit operation

        Args:
            input_data: data with features, target and ids to process
        """

        pass

    def transform(self, input_data: InputData) -> OutputData:
        """ Cut first cut_part from time series\n
            ``new_len = len - int(self.cut_part * (input_values.shape[0]-horizon))``

        Args:
            input_data: data with features, target and ids to process

        Returns:
            output data with cutted time series
        """

        input_data = self._cut_input_data(input_data)

        output_data = self._convert_to_output(input_data,
                                              input_data.features,
                                              data_type=input_data.data_type)
        return output_data

    def transform_for_fit(self, input_data: InputData) -> OutputData:
        """ Cut first cut_part from time series for fit stage
            ``new_len = len - int(self.cut_part * (input_values.shape[0]-horizon))``

        Args:
            input_data: data with features, target and ids to process

        Returns:
            output data with cutted time series
        """

        input_data = self._cut_input_data(input_data, reset_idx=True)

        output_data = self._convert_to_output(input_data,
                                              input_data.features,
                                              data_type=input_data.data_type)
        return output_data

    def _cut_input_data(self, input_data: InputData, reset_idx: bool = False) -> InputData:
        horizon = input_data.task.task_params.forecast_length
        input_copy = copy(input_data)
        input_values = input_copy.features

        cut_len = int(self.cut_part * (input_values.shape[0] - horizon))
        output_values = input_values[cut_len::]

        input_copy.features = output_values
        input_copy.target = output_values

        if reset_idx:
            input_copy.idx = np.arange(cut_len, input_values.shape[0])
        return input_copy


def ts_to_table(idx, time_series: np.array, window_size: int, is_lag: bool = False):
    """Method convert time series to lagged form.

    Args:
        idx: the indices of the time series to convert
        time_series: source time series
        window_size: size of sliding window, which defines lag
        is_lag: is function used for lagged transformation.
            ``False`` needs to convert one dimensional output to lagged form.

    Returns:
        ``updated_idx`` -> clipped indices of time series\n
        ``features_columns`` -> lagged time series feature table
    """
    # Convert data to lagged form
    lagged_dataframe = pd.DataFrame({'t_id': time_series})
    vals = lagged_dataframe['t_id']
    for i in range(1, window_size):
        frames = [lagged_dataframe, vals.shift(i)]
        lagged_dataframe = pd.concat(frames, axis=1)

    # Remove incomplete rows
    lagged_dataframe.dropna(inplace=True)

    transformed = np.array(lagged_dataframe)

    # Generate dataset with features
    features_columns = np.fliplr(transformed)

    if is_lag:
        updated_idx = list(idx[window_size:])
        updated_idx.append(idx[-1])
        updated_idx = np.array(updated_idx)
    else:
        updated_idx = idx[:len(idx) - window_size + 1]

    return updated_idx, features_columns


def _sparse_matrix(logger, features_columns: np.array, n_components_perc=0.5, use_svd=False):
    """Method converts the matrix to sparse form

        Args:
            features_columns: matrix to sparse
            n_components_perc: initial approximation of percent of components to keep
            use_svd: is there need to use :obj:`SVD` method for sparse or use naive method

        Returns:
            reduced dimension matrix

        Notes:
            shape of returned matrix depends on the number of components which includes
            the threshold of explained variance gain
        """

    if not n_components_perc:
        n_components_perc = 0.5

    if use_svd:
        n_components = int(features_columns.shape[1] * n_components_perc)
        # Getting the initial approximation of number of components
        if not n_components:
            n_components = int(features_columns.shape[1] * n_components_perc)
        if n_components >= features_columns.shape[0]:
            n_components = features_columns.shape[0] - 1
        logger.info(f'Initial approximation of number of components set as {n_components}')

        # Forming the first value of explained variance
        components = _get_svd(features_columns, n_components)
    else:
        step = int(1 / n_components_perc)
        indeces_to_stay = np.arange(1, features_columns.shape[1], step)
        components = np.take(features_columns, indeces_to_stay, 1)

    return components


def _get_svd(features_columns: np.array, n_components: int):
    """Method converts the matrix to svd sparse form

    Args:
        features_columns: matrix to sparse
        n_components: number of components to keep

    Returns:
        transformed sparse matrix
    """

    svd = TruncatedSVD(n_components=n_components, n_iter=5, random_state=42)
    svd.fit(features_columns.T)
    components = svd.components_.T
    return components


def prepare_target(all_idx, idx, features_columns: np.array, target, forecast_length: int):
    """Method convert time series to lagged form. Transformation applied
    only for generating target table (time series considering as multi-target
    regression task)

    Args:
        all_idx: all indices in data
        idx: remaining indices after lagged feature table generation
        features_columns: lagged feature table
        target: source time series
        forecast_length: forecast length

    Returns:
        ``updated_idx``, ``updated_features``, ``updated_target``

        .. details:: more information:

            - ``updated_idx`` -> clipped indices of time series
            - ``updated_features`` -> clipped lagged feature table
            - ``updated_target`` -> lagged target table

    """
    # Remove last repeated element
    idx = idx[: -1]

    # Update target (clip first "window size" values)
    row_nums = [list(all_idx).index(i) for i in idx]
    ts_target = target[row_nums]

    # Multi-target transformation
    if forecast_length > 1:
        # Target transformation
        df = pd.DataFrame({'t_id': ts_target})
        vals = df['t_id']
        for i in range(1, forecast_length):
            frames = [df, vals.shift(-i)]
            df = pd.concat(frames, axis=1)

        # Remove incomplete rows
        df.dropna(inplace=True)
        updated_target = np.array(df)

        updated_idx = idx[: -forecast_length + 1]
        updated_features = features_columns[: -forecast_length]
    else:
        # Forecast horizon equals to 1
        updated_idx = idx
        updated_features = features_columns[: -1]
        updated_target = ts_target

    return updated_idx, updated_features, updated_target


def transform_features_and_target_into_lagged(input_data: InputData, forecast_length: int,
                                              window_size: int):
    """Perform lagged transformation firstly on features and secondly on target array

    Args:
        input_data: dataclass with features
        forecast_length: forecast horizon
        window_size: window size for features transformation

    Returns:
        ``new_idx``, ``transformed_cols``, ``new_target``

        .. details:: more information:

            - ``new_idx`` ->
            - ``transformed_cols`` ->
            - ``new_target`` ->

    """

    new_idx, transformed_cols = ts_to_table(idx=input_data.idx,
                                            time_series=input_data.features,
                                            window_size=window_size,
                                            is_lag=True)
    new_idx, transformed_cols, new_target = prepare_target(all_idx=input_data.idx,
                                                           idx=new_idx,
                                                           features_columns=transformed_cols,
                                                           target=input_data.target,
                                                           forecast_length=forecast_length)
    return new_idx, transformed_cols, new_target



import logging
import os
from typing import Optional

import numpy as np

from fedot.core.operations.operation_parameters import OperationParameters
from fedot.utilities.requirements_notificator import warn_requirement

try:
    import tensorflow as tf
except ModuleNotFoundError:
    warn_requirement('tensorflow')
    tf = None

from fedot.core.data.data import InputData, OutputData
from fedot.core.log import LoggerAdapter, default_log
from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import ModelImplementation
from sklearn import preprocessing

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'


def check_input_array(x_train):
    if np.max(x_train) > 1:
        transformed_x_train = x_train.astype("float32") / 255
        transform_flag = True
    else:
        transformed_x_train = x_train
        transform_flag = False

    return transformed_x_train, transform_flag


def create_deep_cnn(input_shape: tuple,
                    num_classes: int):
    model = tf.keras.Sequential(
        [
            tf.keras.layers.InputLayer(input_shape=input_shape),
            tf.keras.layers.Conv2D(32, kernel_size=(3, 3), activation="relu"),
            tf.keras.layers.MaxPooling2D(pool_size=(2, 2)),
            tf.keras.layers.Conv2D(64, kernel_size=(3, 3), activation="relu"),
            tf.keras.layers.MaxPooling2D(pool_size=(2, 2)),
            tf.keras.layers.Conv2D(128, kernel_size=(3, 3), activation="relu"),
            tf.keras.layers.MaxPooling2D(pool_size=(2, 2)),
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dropout(0.5),
            tf.keras.layers.Dense(num_classes, activation="softmax"),
        ]
    )
    return model


def create_simple_cnn(input_shape: tuple,
                      num_classes: int):
    model = tf.keras.Sequential(
        [
            tf.keras.layers.InputLayer(input_shape=input_shape),
            tf.keras.layers.Conv2D(32, kernel_size=(3, 3), activation="relu"),
            tf.keras.layers.MaxPooling2D(pool_size=(2, 2)),
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dropout(0.5),
            tf.keras.layers.Dense(num_classes, activation="softmax"),
        ]
    )

    return model


def create_vgg16(input_shape: tuple,
                 num_classes: int):
    model = tf.keras.applications.vgg16.VGG16(include_top=True,
                                              weights=None,
                                              input_shape=input_shape,
                                              classes=num_classes,
                                              classifier_activation='sigmoid')
    return model


def fit_cnn(train_data: InputData,
            model,
            epochs: int = 10,
            batch_size: int = 128,
            optimizer_params: dict = None,
            logger: Optional[LoggerAdapter] = None):
    x_train, y_train = train_data.features, train_data.target
    transformed_x_train, transform_flag = check_input_array(x_train)

    if logger is None:
        logger = default_log(prefix=__name__)

    if transform_flag:
        logger.debug('Train data set was not scaled. The data was divided by 255.')

    if len(x_train.shape) == 3:
        transformed_x_train = np.expand_dims(x_train, -1)

    if len(train_data.target.shape) < 2:
        le = preprocessing.OneHotEncoder()
        y_train = le.fit_transform(y_train.reshape(-1, 1)).toarray()

    if optimizer_params is None:
        optimizer_params = {'loss': "categorical_crossentropy",
                            'optimizer': "adam",
                            'metrics': ["accuracy"]}

    model.compile(**optimizer_params)
    model.num_classes = train_data.num_classes
    if logger is None:
        logger = default_log(prefix=__name__)

    if logger.logging_level > logging.DEBUG:
        verbose = 0
    else:
        verbose = 2

    if epochs is None:
        logger.warning('The number of training epochs was not set. The selected number of epochs is 10.')

    model.fit(transformed_x_train, y_train, batch_size=batch_size, epochs=epochs,
              validation_split=0.1, verbose=verbose)

    return model


def predict_cnn(trained_model, predict_data: InputData, output_mode: str = 'labels', logger=None) -> OutputData:
    x_test = predict_data.features
    transformed_x_test, transform_flag = check_input_array(x_test)

    if logger is None:
        logger = default_log(prefix=__name__)

    if np.max(transformed_x_test) > 1:
        logger.warning('Test data set was not scaled. The data was divided by 255.')

    if len(x_test.shape) == 3:
        transformed_x_test = np.expand_dims(x_test, -1)

    if output_mode == 'labels':
        prediction = np.round(trained_model.predict(transformed_x_test))
    elif output_mode in ['probs', 'full_probs', 'default']:
        prediction = trained_model.predict(transformed_x_test)
        if trained_model.num_classes < 2:
            logger.error('Data set contain only 1 target class. Please reformat your data.')
            raise NotImplementedError()
        elif trained_model.num_classes == 2 and output_mode != 'full_probs' and len(prediction.shape) > 1:
            prediction = prediction[:, 1]
    else:
        raise ValueError(f'Output model {output_mode} is not supported')
    return prediction


cnn_model_dict = {'deep': create_deep_cnn,
                  'simplified': create_simple_cnn,
                  'vgg16': create_vgg16}


class FedotCNNImplementation(ModelImplementation):
    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)

        default_params = {'log': default_log(prefix=__name__),
                          'epochs': 10,
                          'batch_size': 32,
                          'output_mode': 'labels',
                          'architecture_type': 'simplified',
                          'optimizer_parameters': {'loss': "categorical_crossentropy",
                                                   'optimizer': "adam",
                                                   'metrics': ["accuracy"]}}

        complete_params = {**default_params, **self.params.to_dict()}
        self.params.update(**complete_params)

    def fit(self, train_data):
        """ Method fit model on a dataset

        :param train_data: data to train the model
        """

        # TODO make case for multiclass multioutput task
        # check for multioutput target
        if len(train_data.target.shape) < 2:
            self.classes = np.unique(train_data.target)
        else:
            self.classes = np.arange(train_data.target.shape[1])

        self.model = cnn_model_dict[self.params.get('architecture_type')](input_shape=train_data.features.shape[1:4],
                                                                          num_classes=len(self.classes))

        self.model = fit_cnn(train_data=train_data, model=self.model, epochs=self.params.get('epochs'),
                             batch_size=self.params.get('batch_size'),
                             optimizer_params=self.params.get('optimizer_parameters'), logger=self.params.get('log'))
        return self.model

    def predict(self, input_data):
        """ Method make prediction with labels of classes for predict stage

        :param input_data: data with features to process
        """

        return predict_cnn(trained_model=self.model, predict_data=input_data,
                           output_mode='labels', logger=self.params['log'])

    def predict_proba(self, input_data):
        """ Method make prediction with probabilities of classes

        :param input_data: data with features to process
        """

        return predict_cnn(trained_model=self.model, predict_data=input_data, output_mode='probs')

    @property
    def classes_(self):
        return self.classes

    def __deepcopy__(self, memo=None):
        clone_model = tf.keras.models.clone_model(self.model)
        clone_model.compile(optimizer=self.model.optimizer, loss=self.model.loss, metrics=self.model.metrics)
        clone_model.set_weights(self.model.get_weights())
        return clone_model



from copy import deepcopy
from typing import Optional, Union

import numpy as np
from sklearn.preprocessing import OneHotEncoder, LabelEncoder

from fedot.core.data.data import InputData, OutputData
from fedot.core.data.data_preprocessing import find_categorical_columns
from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import \
    DataOperationImplementation
from fedot.core.operations.operation_parameters import OperationParameters


class OneHotEncodingImplementation(DataOperationImplementation):
    """ Class for automatic categorical data detection and one hot encoding """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        default_params = {
            'handle_unknown': 'ignore'
        }
        self.encoder = OneHotEncoder(**{**default_params, **self.params.to_dict()})
        self.categorical_ids = None
        self.non_categorical_ids = None

    def fit(self, input_data: InputData):
        """ Method for fit encoder with automatic determination of categorical features

        :param input_data: data with features, target and ids for encoder training
        :return encoder: trained encoder (optional output)
        """
        features = input_data.features
        features_types = input_data.supplementary_data.column_types.get('features')
        categorical_ids, non_categorical_ids = find_categorical_columns(features,
                                                                        features_types)

        # Indices of columns with categorical and non-categorical features
        self.categorical_ids = categorical_ids
        self.non_categorical_ids = non_categorical_ids

        # If there are categorical features - process it
        if self.categorical_ids:
            updated_cat_features = np.array(features[:, self.categorical_ids], dtype=str)
            self.encoder.fit(updated_cat_features)

        return self.encoder

    def transform(self, input_data: InputData) -> OutputData:
        """
        The method that transforms the categorical features in the original
        dataset, but does not affect the rest features. Applicable during predict stage

        :param input_data: data with features, target and ids for transformation
        :return output_data: output data with transformed features table
        """
        copied_data = deepcopy(input_data)

        features = copied_data.features
        if not self.categorical_ids:
            # If there are no categorical features in the table
            transformed_features = features
        else:
            # If categorical features are exists
            transformed_features = self._apply_one_hot_encoding(features)

        # Update features
        output_data = self._convert_to_output(copied_data,
                                              transformed_features)
        self._update_column_types(output_data)
        return output_data

    def _update_column_types(self, output_data: OutputData):
        """ Update column types after encoding. Categorical columns becomes integer with extension """
        if self.categorical_ids:
            # There are categorical features in the table
            col_types = output_data.supplementary_data.column_types['features']
            numerical_columns = [t_name for t_name in col_types if 'str' not in t_name]

            # Calculate new binary columns number after encoding
            encoded_columns_number = output_data.predict.shape[1] - len(numerical_columns)
            numerical_columns.extend([str(int)] * encoded_columns_number)

            output_data.supplementary_data.column_types['features'] = numerical_columns

    def _apply_one_hot_encoding(self, features: np.array) -> np.array:
        """
        The method creates a table based on categorical and real features after One Hot Encoding transformation

        :param features: tabular data for processing
        :return transformed_features: transformed features table
        """

        categorical_features = np.array(features[:, self.categorical_ids])
        transformed_categorical = self.encoder.transform(categorical_features).toarray()

        # If there are non-categorical features in the data
        if not self.non_categorical_ids:
            transformed_features = transformed_categorical
        else:
            # Stack transformed categorical and non-categorical data
            non_categorical_features = np.array(features[:, self.non_categorical_ids])
            frames = (non_categorical_features, transformed_categorical)
            transformed_features = np.hstack(frames)

        return transformed_features


class LabelEncodingImplementation(DataOperationImplementation):
    """ Class for categorical features encoding based on LabelEncoding """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        # LabelEncoder has no parameters
        self.encoders = {}
        self.categorical_ids = None
        self.non_categorical_ids = None

    def fit(self, input_data: InputData):
        features_types = input_data.supplementary_data.column_types.get('features')
        self.categorical_ids, self.non_categorical_ids = find_categorical_columns(input_data.features,
                                                                                  features_types)

        # If there are categorical features - process it
        if self.categorical_ids:
            # For every categorical feature - perform encoding
            self._fit_label_encoders(input_data)
        return self.encoders

    def transform(self, input_data: InputData) -> OutputData:
        """ Apply LabelEncoder on categorical features and doesn't process float or int ones
        Applicable during predict stage
        """
        copied_data = deepcopy(input_data)
        if self.categorical_ids:
            # If categorical features are exists - transform them inplace in InputData
            for categorical_id in self.categorical_ids:
                categorical_column = input_data.features[:, categorical_id]

                # Converting into string - so nans becomes marked as 'nan'
                categorical_column = categorical_column.astype(str)
                gap_ids = np.ravel(np.argwhere(categorical_column == 'nan'))

                transformed = self._apply_label_encoder(categorical_column, categorical_id, gap_ids)
                copied_data.features[:, categorical_id] = transformed

        output_data = self._convert_to_output(input_data,
                                              copied_data.features)

        self._update_column_types(output_data)
        return output_data

    def _update_column_types(self, output_data: OutputData):
        """ Update column types after encoding. Categorical becomes integer """
        if self.categorical_ids:
            # Categorical features were in the dataset
            col_types = output_data.supplementary_data.column_types['features']
            for categorical_id in self.categorical_ids:
                col_types[categorical_id] = str(int)

            output_data.supplementary_data.column_types['features'] = col_types

    def _fit_label_encoders(self, input_data: InputData):
        """ Fit LabelEncoder for every categorical column in the dataset """
        for categorical_id in self.categorical_ids:
            categorical_column = input_data.features[:, categorical_id]
            le = LabelEncoder()
            le.fit(categorical_column)

            self.encoders.update({categorical_id: le})

    def _apply_label_encoder(self, categorical_column: np.array, categorical_id: int,
                             gap_ids: Union[np.array, None]) -> np.array:
        """ Apply fitted LabelEncoder for column transformation

        :param categorical_column: numpy array with categorical features
        :param categorical_id: index of current categorical column
        :param gap_ids: indices of gap elements in array
        """
        column_encoder = self.encoders[categorical_id]
        encoder_classes = list(column_encoder.classes_)

        # If the column contains categories not previously encountered
        for label in list(set(categorical_column)):
            if label not in encoder_classes:
                encoder_classes.append(label)

        # Extent encoder classes
        column_encoder.classes_ = np.array(encoder_classes)

        transformed_column = column_encoder.transform(categorical_column)
        if len(gap_ids) > 0:
            # Store np.nan values
            transformed_column = transformed_column.astype(object)
            transformed_column[gap_ids] = np.nan

        return transformed_column

    def get_params(self) -> OperationParameters:
        """ Due to LabelEncoder has no parameters - return empty set """
        return OperationParameters()



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.metrics import mean_absolute_error, mean_squared_error, median_absolute_error

from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.utils import fedot_project_root
from fedot.utilities.ts_gapfilling import ModelGapFiller


def print_metrics(dataframe):
    """
    The function displays 3 metrics: Mean absolute error,
    Root mean squared error and Median absolute error

    :param dataframe: dataframe with columns 'date','temperature','ridge','composite',
    'with_gap'
    """

    gap_array = np.array(dataframe['with_gap'])
    gap_ids = np.argwhere(gap_array == -100.0)

    actual = np.array(dataframe['temperature'])[gap_ids]
    ridge_predicted = np.array(dataframe['ridge'])[gap_ids]
    composite_predicted = np.array(dataframe['composite'])[gap_ids]

    model_labels = ['Inverted ridge regression', 'Composite model']
    for predicted, model_label in zip(
            [ridge_predicted, composite_predicted], model_labels):
        print(f"{model_label}")

        mae_metric = mean_absolute_error(actual, predicted)
        print(f"Mean absolute error - {mae_metric:.2f}")

        rmse_metric = (mean_squared_error(actual, predicted)) ** 0.5
        print(f"Root mean squared error - {rmse_metric:.2f}")

        median_ae_metric = median_absolute_error(actual, predicted)
        print(f"Median absolute error - {median_ae_metric:.2f} \n")


def plot_result(dataframe):
    """
    The function draws a graph based on the dataframe

    :param dataframe: dataframe with columns 'date','temperature','ridge','composite',
    'with_gap'
    """

    gap_array = np.array(dataframe['with_gap'])
    masked_array = np.ma.masked_where(gap_array == -100.0, gap_array)

    plt.plot(dataframe['date'], dataframe['temperature'], c='blue',
             alpha=0.5, label='Actual values', linewidth=1)
    plt.plot(dataframe['date'], dataframe['ridge'], c='orange',
             alpha=0.8, label='Bi-directional ridge gap-filling', linewidth=1)
    plt.plot(dataframe['date'], dataframe['composite'], c='red',
             alpha=0.8, label='Composite gap-filling', linewidth=1)
    plt.plot(dataframe['date'], masked_array, c='blue')
    plt.grid()
    plt.legend()
    plt.show()


def get_composite_pipeline():
    """
    The function returns prepared pipeline of 5 models

    :return: Pipeline object
    """

    node_1 = PrimaryNode('lagged')
    node_1.parameters = {'window_size': 150}
    node_2 = PrimaryNode('lagged')
    node_2.parameters = {'window_size': 100}
    node_linear_1 = SecondaryNode('linear', nodes_from=[node_1])
    node_linear_2 = SecondaryNode('linear', nodes_from=[node_2])

    node_final = SecondaryNode('ridge', nodes_from=[node_linear_1,
                                                    node_linear_2])
    pipeline = Pipeline(node_final)
    return pipeline


def get_simple_pipeline():
    """ Function returns simple pipeline """
    node_lagged = PrimaryNode('lagged')
    node_lagged.parameters = {'window_size': 150}
    node_ridge = SecondaryNode('ridge', nodes_from=[node_lagged])
    ridge_pipeline = Pipeline(node_ridge)
    return ridge_pipeline


def run_gapfilling_case(file_path):
    """
    The function runs an example of filling in gaps in a time series with
    air temperature. Real data case.

    :param file_path: path to the file
    :return: pandas dataframe with columns 'date','with_gap','ridge',
    'composite','temperature'
    """

    # Load dataframe
    full_path = os.path.join(str(fedot_project_root()), file_path)
    dataframe = pd.read_csv(full_path)
    dataframe['date'] = pd.to_datetime(dataframe['date'])

    # Filling in gaps based on inverted ridge regression model
    ridge_pipeline = get_simple_pipeline()
    ridge_gapfiller = ModelGapFiller(gap_value=-100.0,
                                     pipeline=ridge_pipeline)
    with_gap_array = np.array(dataframe['with_gap'])
    without_gap_arr_ridge = ridge_gapfiller.forward_inverse_filling(with_gap_array)
    dataframe['ridge'] = without_gap_arr_ridge

    # Filling in gaps based on a pipeline of 5 models
    composite_pipeline = get_composite_pipeline()
    composite_gapfiller = ModelGapFiller(gap_value=-100.0,
                                         pipeline=composite_pipeline)
    without_gap_composite = composite_gapfiller.forward_filling(with_gap_array)
    dataframe['composite'] = without_gap_composite
    return dataframe


# Example of using the algorithm to fill in gaps in a time series
# The data is daily air temperature values from the weather station
if __name__ == '__main__':
    dataframe = run_gapfilling_case('cases/data/gapfilling/ts_temperature_gapfilling.csv')

    # Display metrics
    print_metrics(dataframe)

    # Visualise predictions
    plot_result(dataframe)



import warnings

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from sklearn.metrics import mean_squared_error, mean_absolute_error

from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.ts_wrappers import in_sample_ts_forecast
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import TaskTypesEnum, Task, TsForecastingParams

warnings.filterwarnings('ignore')
np.random.seed(2020)


def get_refinement_pipeline_with_polyfit():
    """ Create 4-level pipeline with decompose operation """

    node_polyfit = PrimaryNode('polyfit')
    node_polyfit.parameters = {'degree': 2}
    node_lagged = PrimaryNode('lagged')
    node_decompose = SecondaryNode('decompose', nodes_from=[node_lagged, node_polyfit])
    node_dtreg = SecondaryNode('dtreg', nodes_from=[node_decompose])
    node_dtreg.parameters = {'max_depth': 3}

    # Pipelines with different outputs
    pipeline_with_decompose_finish = Pipeline(node_dtreg)
    pipeline_with_main_finish = Pipeline(node_polyfit)

    # Combining branches with different targets (T and T_decomposed)
    final_node = SecondaryNode('ridge', nodes_from=[node_polyfit, node_dtreg])

    pipeline = Pipeline(final_node)
    return pipeline_with_main_finish, pipeline_with_decompose_finish, pipeline


def get_refinement_pipeline(lagged):
    """ Create 4-level pipeline with decompose operation """

    node_lagged = PrimaryNode('lagged')
    node_lagged.parameters = {'window_size': lagged}
    node_lasso = SecondaryNode('lasso', nodes_from=[node_lagged])
    node_decompose = SecondaryNode('decompose', nodes_from=[node_lagged, node_lasso])
    node_dtreg = SecondaryNode('dtreg', nodes_from=[node_decompose])
    node_dtreg.parameters = {'max_depth': 3}

    # Pipelines with different outputs
    pipeline_with_decompose_finish = Pipeline(node_dtreg)
    pipeline_with_main_finish = Pipeline(node_lasso)

    # Combining branches with different targets (T and T_decomposed)
    final_node = SecondaryNode('ridge', nodes_from=[node_lasso, node_dtreg])

    pipeline = Pipeline(final_node)
    return pipeline_with_main_finish, pipeline_with_decompose_finish, pipeline


def get_non_refinement_pipeline(lagged):
    """ Create 4-level pipeline without decompose operation """

    node_lagged = PrimaryNode('lagged')
    node_lagged.parameters = {'window_size': lagged}
    node_lasso = SecondaryNode('lasso', nodes_from=[node_lagged])
    node_dtreg = SecondaryNode('dtreg', nodes_from=[node_lagged])
    node_dtreg.parameters = {'max_depth': 3}
    final_node = SecondaryNode('ridge', nodes_from=[node_lasso, node_dtreg])

    pipeline = Pipeline(final_node)
    return pipeline


def in_sample_fit_predict(pipeline, train_input, predict_input, horizon) -> np.array:
    """ Fit pipeline and make predictions (in-sample forecasting) """
    pipeline.fit(train_input)

    predicted_main = in_sample_ts_forecast(pipeline=pipeline,
                                           input_data=predict_input,
                                           horizon=horizon)
    return predicted_main


def display_metrics(test_part, predicted_values, pipeline_name):
    mse = mean_squared_error(test_part.target, predicted_values, squared=False)
    mae = mean_absolute_error(test_part.target, predicted_values)
    print(f'RMSE {pipeline_name} - {mse:.4f}')
    print(f'MAE {pipeline_name} - {mae:.4f}\n')


def run_refinement_forecast(path_to_file, len_forecast=100, lagged=150,
                            validation_blocks=3, vis_with_decompose=True):
    """ Function launch example with experimental features of the FEDOT framework

    :param path_to_file: path to the csv file
    :param len_forecast: forecast length
    :param lagged: window size for lagged transformation
    :param validation_blocks: amount of parts for time series validation
    :param vis_with_decompose: visualise part of main forecast
    """

    # Read dataframe
    df = pd.read_csv(path_to_file)
    time_series = np.array(df['value'])
    idx = np.array(pd.to_datetime(df['datetime']))
    # 3 folds for validation
    horizon = len_forecast * validation_blocks

    # Get pipeline with decomposing operation
    pipeline_with_main_finish, pipeline_with_decompose_finish, pipeline = get_refinement_pipeline(lagged)
    # Get simple pipeline without decomposing operation
    simple_pipeline = get_non_refinement_pipeline(lagged)

    train_input, predict_input = train_test_data_setup(
        InputData(idx=idx,
                  features=time_series,
                  target=time_series,
                  task=Task(TaskTypesEnum.ts_forecasting,
                            TsForecastingParams(
                                forecast_length=horizon)),
                  data_type=DataTypesEnum.ts))

    # Forecast of pipeline with decomposition
    predicted_values = in_sample_fit_predict(pipeline, train_input,
                                             predict_input, horizon)
    display_metrics(predict_input, predicted_values, pipeline_name='with decomposition')

    # Range for visualisation
    ids_for_test = idx[range(len(train_input.target), len(time_series))]
    plt.plot(idx, time_series, label='Actual time series')
    plt.plot(ids_for_test, predicted_values, label='With decomposition')

    if vis_with_decompose:
        # Forecast of first model in the pipeline
        predicted_main = in_sample_fit_predict(pipeline_with_main_finish, train_input,
                                               predict_input, horizon)
        # Forecast for residuals
        predicted_decompose = in_sample_fit_predict(pipeline_with_decompose_finish, train_input,
                                                    predict_input, horizon)

        plt.plot(ids_for_test, predicted_main, label='Main branch forecast')
        plt.plot(ids_for_test, predicted_decompose, label='Residual branch forecast')
    else:
        predicted_simple = in_sample_fit_predict(simple_pipeline, train_input,
                                                 predict_input, horizon)
        plt.plot(ids_for_test, predicted_simple, label='Non decomposition')
        display_metrics(predict_input, predicted_simple, pipeline_name='without decomposition')

    i = len(train_input.target)
    for _ in range(0, validation_blocks):
        deviation = np.std(predicted_values)
        plt.plot([idx[i], idx[i]], [min(predicted_values) - deviation, max(predicted_values) + deviation],
                 c='black', linewidth=1)
        i += len_forecast
    plt.grid()
    plt.legend()
    plt.show()


if __name__ == '__main__':
    path = '../../../cases/data/time_series/economic_data.csv'
    run_refinement_forecast(path, len_forecast=50, validation_blocks=5,
                            lagged=50, vis_with_decompose=True)



import os
import h2o


class TPOTRegressionSerializationWrapper:
    """ Wrapper to serialize tpot algorithms.
    Can be used for classification, multioutput regression and time series forecasting"""
    def __init__(self, estimators):
        self._estimators = estimators

    def get_estimators(self):
        return self._estimators


class H2OSerializationWrapper:
    """ Wrapper to serialize h2o algorithms.
    Can be used for classification, multioutput regression and time series forecasting.
    Unfortunately there are no support for all types of h2o pipelines (for this version)"""
    def __init__(self, estimators):
        self._estimators = estimators

    @classmethod
    def load_operation(cls, path_global):
        models = []
        for path in os.listdir(path_global):
            path = os.path.join(path_global, path)
            imported_model = h2o.import_mojo(path)
            models.append(imported_model)
        return H2OSerializationWrapper(models)

    def get_estimators(self):
        return self._estimators

    def save_operation(self, path, operation_id):
        path = os.path.join(path, f'h2o_{operation_id}')
        count = 0
        for model in self._estimators:
            model_name = f"{count}.zip"
            model.save_mojo(os.path.join(path, model_name))
            count += 1
        return path



import os
import shutil
from pathlib import Path
from typing import Optional, Tuple, Union

from fedot.core.data.data import InputData
from fedot.core.log import LoggerAdapter, default_log
from fedot.core.optimisers.opt_history_objects.opt_history import OptHistory
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.utils import default_fedot_data_dir

DEFAULT_PATH = Path(default_fedot_data_dir())
DEFAULT_PROJECTS_PATH = DEFAULT_PATH.joinpath('projects')


def export_project_to_zip(zip_name: Union[str, Path], pipeline: Pipeline, train_data: InputData, test_data: InputData,
                          opt_history: Optional[OptHistory] = None, log_file_name: str = 'log.txt'):
    """
    Convert pipeline to JSON, data to csv, compress them to zip
    archive and save to 'DEFAULT_PROJECTS_PATH/projects' with logs.

    :param zip_name: absolute or relative path the zip file with exported project
    :param pipeline: pipeline object to export
    :param train_data: train InputData object to export
    :param test_data: test InputData object to export
    :param opt_history: history of model optimisation to export (if available)
    :param log_file_name: name of the file with log to export
    """

    log = default_log(prefix='fedot.utilities.project_import_export')
    absolute_folder_path, absolute_zip_path, folder_name, zip_name = _prepare_paths(zip_name)
    _check_for_existing_project(absolute_folder_path, absolute_zip_path)

    # Converts python objects to files for compression
    pipeline_path = os.path.join(absolute_folder_path, 'pipeline', 'pipeline.json')
    pipeline.save(pipeline_path, is_datetime_in_path=False, create_subdir=False)
    train_data.to_csv(absolute_folder_path.joinpath('train_data.csv'))
    test_data.to_csv(absolute_folder_path.joinpath('test_data.csv'))
    if opt_history is not None:
        opt_history.save(Path(absolute_folder_path, 'opt_history.json'))

    _copy_log_file(log_file_name, absolute_folder_path)

    shutil.make_archive(base_name=absolute_zip_path.with_suffix(''), format='zip', root_dir=absolute_folder_path)
    folder_to_delete = DEFAULT_PROJECTS_PATH.joinpath(
        absolute_folder_path.relative_to(DEFAULT_PROJECTS_PATH).parts[0])
    shutil.rmtree(folder_to_delete)

    log.info(f'The exported project was saved on the path: {absolute_zip_path}')


def import_project_from_zip(zip_path: str) -> Tuple[Pipeline, InputData, InputData, OptHistory]:
    """
    Unzipping zip file. Zip file should contain:
    - pipeline.json: json performance,
    - train_data.csv: csv with first line which contains task_type and data_type of train InputData object,
    - test_data.csv: csv with first line which contains task_type and data_type of test InputData object.

    Created Pipeline and InputData objects. Ready to work with it.

    :param zip_path: path to zip archive
    :return imported classes
    """
    log = default_log(prefix='fedot.utilities.project_import_export')

    folder_path, absolute_zip_path, _, zip_name = _prepare_paths(zip_path)

    zip_path = _check_zip_path(absolute_zip_path, log)

    if folder_path.exists():
        # ensure temporary folder is clear
        shutil.rmtree(folder_path)
    shutil.unpack_archive(zip_path, folder_path)

    message = f'The project "{zip_name}" was unpacked to the "{folder_path}".'
    log.info(message)
    pipeline, train_data, test_data, opt_history = [None] * 4
    for root, dirs, files in os.walk(folder_path):
        for file in files:
            if file == 'pipeline.json':
                pipeline = Pipeline.from_serialized(os.path.join(root, file))
            elif file == 'train_data.csv':
                train_data = InputData.from_csv(os.path.join(root, file))
            elif file == 'test_data.csv':
                test_data = InputData.from_csv(os.path.join(root, file))
            elif file == 'opt_history.json':
                opt_history = OptHistory.load(os.path.join(root, file))

    shutil.rmtree(folder_path)
    return pipeline, train_data, test_data, opt_history


def _check_zip_path(zip_path: Path, log: LoggerAdapter) -> Path:
    """Check 'zip_path' for correctness."""

    zip_path = zip_path.with_suffix('.zip')

    if not zip_path.exists():
        message = f'File with the path "{zip_path}" could not be found.'
        log.error(message)
        raise FileNotFoundError(message)
    return zip_path


def _copy_log_file(log_file_name: Optional[str], absolute_folder_path: Path):
    """Copy log file to folder which will be compressed."""

    if log_file_name is None:
        return

    log_file_name = Path(log_file_name)

    if not log_file_name.is_absolute():
        log_file_name = DEFAULT_PATH.joinpath(log_file_name).resolve()

    if log_file_name.exists():
        shutil.copy2(log_file_name,
                     absolute_folder_path.joinpath(log_file_name.stem).with_suffix(log_file_name.suffix))


def _prepare_paths(zip_path: Union[str, Path]) -> Tuple[Path, Path, Path, Path]:
    """Prepared paths and names: absolute folder path, absolute zip path, folder name, zip name."""

    zip_path = Path(zip_path)
    if Path(zip_path).suffix:
        folder_name = Path(zip_path.with_suffix('').name)
    else:
        folder_name = Path(zip_path.name)
        zip_path = zip_path.with_suffix('.zip')

    absolute_folder_path = DEFAULT_PROJECTS_PATH.joinpath(folder_name)
    if not zip_path.is_absolute():
        absolute_zip_path = Path.cwd().joinpath(zip_path)
    else:
        absolute_zip_path = zip_path
        zip_path = Path(zip_path.name)
    return absolute_folder_path, absolute_zip_path, folder_name, zip_path


def _check_for_existing_project(absolute_folder_path: Path, zip_path: Path):
    """Check for existing folder and zipfile of project. Create it, if it is no exists."""
    if zip_path.exists():
        message = f'Zipfile with the name "{zip_path}" exists.'
        raise FileExistsError(message)

    if absolute_folder_path.exists():
        message = f'Project with the name "{absolute_folder_path}" exists.'
        raise FileExistsError(message)
    else:
        absolute_folder_path.mkdir(parents=True)



import os
import sqlite3
from contextlib import closing
from pathlib import Path
from typing import Optional, Sequence, Tuple

import psutil

from fedot.core.utils import default_fedot_data_dir


class BaseCacheDB:
    """
    Base class for caching in database.
    Includes low-level idea of caching data using relational database.

    :param main_table: table to store into or load from
    :param cache_folder: path to the place where cache files should be stored.
    :param use_stats: bool indicating if it is needed to use cache performance dict
    :param stats_keys: sequence of keys for supporting cache effectiveness
    """

    def __init__(self, main_table: str = 'default', cache_folder: Optional[str] = None, use_stats: bool = False,
                 stats_keys: Sequence = ('default_hit', 'default_total')):
        self._main_table = main_table
        self._db_suffix = f'.{main_table}_db'
        if cache_folder is None:
            self.db_path = Path(default_fedot_data_dir())
            self._del_prev_temps()
        else:
            self.db_path = Path(cache_folder)
        self.db_path = self.db_path.joinpath(f'cache_{os.getpid()}').with_suffix(self._db_suffix)

        self._eff_table = 'effectiveness'
        self.use_stats = use_stats
        self._effectiveness_keys = stats_keys
        self._init_eff()

    def get_effectiveness(self) -> Optional[Tuple[int, int, int, int]]:
        """
        Returns effectiveness of the cache in case of enabled `use_stats`, None instead.
        """
        if self.use_stats:
            with closing(sqlite3.connect(self.db_path)) as conn:
                with conn:
                    cur = conn.cursor()
                    cur.execute(f'SELECT {",".join(self._effectiveness_keys)} FROM {self._eff_table};')
                    return cur.fetchone()

    def get_effectiveness_keys(self) -> Sequence:
        """
        Returns all cache effectiveness keys.
        """
        return self._effectiveness_keys

    def reset(self):
        """
        Drops all scores from working table and resets efficiency table values to zero.
        """
        with closing(sqlite3.connect(self.db_path)) as conn:
            with conn:
                cur = conn.cursor()
                if self.use_stats:
                    self._reset_eff(cur)
                self._reset_main(cur)

    def _init_eff(self):
        """
        Initializes effectiveness table.
        """
        if self.use_stats:
            with closing(sqlite3.connect(self.db_path)) as conn:
                with conn:
                    cur = conn.cursor()
                    eff_type = ' INTEGER DEFAULT 0'
                    fields = f'{eff_type},'.join(self._effectiveness_keys) + eff_type
                    cur.execute((
                        f'CREATE TABLE IF NOT EXISTS {self._eff_table} ('
                        'id INTEGER PRIMARY KEY CHECK (id = 1),'
                        f'{fields}'
                        ');'
                    ))
                    cur.execute(f'INSERT OR IGNORE INTO {self._eff_table} DEFAULT VALUES;')

    def _del_prev_temps(self):
        """
        Deletes previously generated unused DB files.
        """
        for file in self.db_path.glob(f'cache_*{self._db_suffix}'):
            try:
                pid = int(file.stem.split('_')[-1])
            except ValueError:
                pid = -1  # old format cache name, remove this line somewhere in the future
            if pid not in psutil.pids():
                try:
                    file.unlink()
                except FileNotFoundError:
                    pass  # it means another process have already killed it
                except PermissionError:
                    pass  # the same

    def _inc_eff(self, cur: sqlite3.Cursor, col: str, inc_val: int = 1):
        """
        Increases `col` score in efficiency table by `inc_val`.

        :param cur: cursor with already installed DB connection
        :param col: column of efficiency table to increase
        :param inc_val: value to increase column
        """
        cur.execute(f'UPDATE {self._eff_table} SET {col} = {col} + {inc_val};')

    def _reset_eff(self, cur: sqlite3.Cursor):
        """
        Resets efficiency table scores to zero.

        :param cur: cursor with already installed DB connection
        """
        cur.execute(f'DELETE FROM {self._eff_table};')
        cur.execute(f'INSERT INTO {self._eff_table} DEFAULT VALUES;')

    def _reset_main(self, cur: sqlite3.Cursor):
        """
        Drops all scores from working table.

        :param cur: cursor with already installed DB connection
        """
        cur.execute(f'DELETE FROM {self._main_table};')

    def __len__(self):
        with closing(sqlite3.connect(self.db_path)) as conn:
            with conn:
                cur = conn.cursor()
                cur.execute(f'SELECT id FROM {self._main_table};')
                all_rows = cur.fetchall()
                return len(all_rows)



from typing import List, Optional, TYPE_CHECKING, Union

from fedot.core.caching.base_cache import BaseCache
from fedot.core.caching.pipelines_cache_db import OperationsCacheDB
from fedot.core.pipelines.node import Node
from fedot.core.utilities.data_structures import ensure_wrapped_in_sequence
from fedot.utilities.debug import is_test_session

if TYPE_CHECKING:
    from fedot.core.pipelines.pipeline import Pipeline


class OperationsCache(BaseCache):
    """
    Stores/loads nodes `fitted_operation` field to increase performance of calculations.

    :param cache_folder: path to the place where cache files should be stored.
    """

    def __init__(self, cache_folder: Optional[str] = None):
        super().__init__(OperationsCacheDB(cache_folder))

    def save_nodes(self, nodes: Union[Node, List[Node]], fold_id: Optional[int] = None):
        """
        :param nodes: node/nodes to be cached
        :param fold_id: optional part of cache item UID
                            (can be used to specify the number of CV fold)
        """
        try:
            mapped = [
                (_get_structural_id(node, fold_id), node.fitted_operation)
                for node in ensure_wrapped_in_sequence(nodes)
                if node.fitted_operation is not None
            ]
            self._db.add_operations(mapped)
        except Exception as ex:
            self.log.warning(f'Nodes can not be saved: {ex}. Continue')
            if is_test_session():
                raise ex

    def save_pipeline(self, pipeline: 'Pipeline', fold_id: Optional[int] = None):
        """
        :param pipeline: pipeline to be cached
        :param fold_id: optional part of cache item UID
                            (can be used to specify the number of CV fold)
        """
        self.save_nodes(pipeline.nodes, fold_id)

    def try_load_nodes(self, nodes: Union[Node, List[Node]], fold_id: Optional[int] = None):
        """
        :param nodes: nodes which fitted state should be loaded from cache
        :param fold_id: optional part of cache item UID
                            (can be used to specify the number of CV fold)
        """
        try:
            nodes_lst = ensure_wrapped_in_sequence(nodes)
            structural_ids = [_get_structural_id(node, fold_id) for node in nodes_lst]
            cached_ops = self._db.get_operations(structural_ids)
            for idx, cached_op in enumerate(cached_ops):
                if cached_op is not None:
                    nodes_lst[idx].fitted_operation = cached_op
                else:
                    nodes_lst[idx].fitted_operation = None
        except Exception as ex:
            self.log.warning(f'Cache can not be loaded: {ex}. Continue.')
            if is_test_session():
                raise ex

    def try_load_into_pipeline(self, pipeline: 'Pipeline', fold_id: Optional[int] = None):
        """
        :param pipeline: pipeline for loading into from cache
        :param fold_id: optional part of cache item UID (number of the CV fold)
        """
        self.try_load_nodes(pipeline.nodes, fold_id)


def _get_structural_id(node: Node, fold_id: Optional[int] = None) -> str:
    """
    Gets unique id from node.

    :param node: node to get uid from
    :param fold_id: fold number to fit data
    :return structural_id: unique node identificator
    """
    structural_id = node.descriptive_id
    structural_id += f'_{fold_id}' if fold_id is not None else ''
    return structural_id



from random import choice, random
from typing import List

from fedot.core.optimisers.gp_comp.operators.mutation import Mutation
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.tuning.hyperparams import ParametersChanger
from fedot.core.repository.operation_types_repository import OperationTypesRepository
from fedot.core.repository.tasks import TaskTypesEnum


def parameter_change_mutation(pipeline: Pipeline, requirements, params, opt_params, **kwargs) -> Pipeline:
    """
    This type of mutation is passed over all nodes and changes
    hyperparameters of the operations with probability - 'node mutation probability'
    which is initialised inside the function
    """
    node_mutation_probability = Mutation.get_mutation_prob(mut_id=opt_params.mutation_strength,
                                                           node=pipeline.root_node)
    for node in pipeline.nodes:
        if random() < node_mutation_probability:
            operation_name = node.operation.operation_type
            current_params = node.parameters

            # Perform specific change for particular parameter
            changer = ParametersChanger(operation_name, current_params)
            try:
                new_params = changer.get_new_operation_params()
                if new_params is not None:
                    node.parameters = new_params
            except Exception as ex:
                pipeline.log.error(ex)
    return pipeline


def boosting_mutation(pipeline: Pipeline, requirements, params, **kwargs) -> Pipeline:
    """ This type of mutation adds the additional 'boosting' cascade to the existing pipeline """

    # TODO: refactor next line to get task_type more obviously
    task_type = params.advisor.task.task_type
    decompose_operations = OperationTypesRepository('data_operation').suitable_operation(
        task_type=task_type, tags=['decompose'])
    decompose_operation = decompose_operations[0]

    existing_pipeline = pipeline

    all_data_operations = OperationTypesRepository('data_operation').suitable_operation(
        task_type=task_type)
    preprocessing_primary_nodes = [n for n in existing_pipeline.nodes if str(n) in all_data_operations]

    if len(preprocessing_primary_nodes) > 0:
        data_source = preprocessing_primary_nodes[0]
    else:
        if task_type == TaskTypesEnum.ts_forecasting:
            data_source = PrimaryNode('simple_imputation')
        else:
            data_source = PrimaryNode('scaling')

    decompose_parents = [existing_pipeline.root_node, data_source]

    boosting_model_candidates = requirements.secondary
    if task_type == TaskTypesEnum.classification:
        # the regression models are required
        boosting_model_candidates = \
            OperationTypesRepository('model').suitable_operation(
                task_type=TaskTypesEnum.regression, forbidden_tags=['non_lagged'])
        if not boosting_model_candidates:
            return pipeline

    new_model = choose_new_model(boosting_model_candidates)

    if task_type == TaskTypesEnum.ts_forecasting:
        non_lagged_ts_models = OperationTypesRepository('model').operations_with_tag(['non_lagged'])
        is_non_lagged_ts_models_in_node = \
            str(existing_pipeline.root_node) in non_lagged_ts_models

        if is_non_lagged_ts_models_in_node:
            # if additional lagged node is required
            lagged_node = SecondaryNode('lagged', nodes_from=[data_source])
            decompose_parents = [existing_pipeline.root_node, lagged_node]

    node_decompose = SecondaryNode(decompose_operation, nodes_from=decompose_parents)

    node_boost = SecondaryNode(new_model, nodes_from=[node_decompose])

    node_final = SecondaryNode(choice(requirements.secondary),
                               nodes_from=[existing_pipeline.root_node, node_boost])
    pipeline = Pipeline(node_final)
    return pipeline


def choose_new_model(boosting_model_candidates: List[str]) -> str:
    """ Since 'linear' and 'dtreg' operations are suitable for solving the problem
    and they are simpler than others, they are preferred """

    if 'linear' in boosting_model_candidates:
        new_model = 'linear'
    elif 'dtreg' in boosting_model_candidates:
        new_model = 'dtreg'
    else:
        new_model = choice(boosting_model_candidates)
    return new_model



from pathlib import Path

from fedot.core.pipelines.adapters import PipelineAdapter
from fedot.core.optimisers.opt_history_objects.opt_history import OptHistory
from fedot.core.utils import fedot_project_root
from fedot.core.visualisation.pipeline_specific_visuals import PipelineHistoryVisualizer


def run_pipeline_and_history_visualization():
    """ The function runs visualization of the composing history and the best pipeline. """
    # Gather pipeline and history.
    history = OptHistory.load(Path(fedot_project_root(), 'examples', 'data', 'histories', 'scoring_case_history.json'))
    pipeline = PipelineAdapter().restore(history.individuals[-1][-1].graph)
    # Show visualizations.
    pipeline.show()
    history_visualizer = PipelineHistoryVisualizer(history)
    history_visualizer.fitness_line()
    history_visualizer.fitness_box(best_fraction=0.5)
    history_visualizer.operations_kde()
    history_visualizer.operations_animated_bar(save_path='example_animation.gif', show_fitness=True)
    history_visualizer.fitness_line_interactive()


if __name__ == '__main__':
    run_pipeline_and_history_visualization()



import pickle
import sqlite3
from contextlib import closing
from os import getpid
from typing import List, Optional, Tuple, TypeVar

from fedot.core.caching.base_cache_db import BaseCacheDB
from fedot.core.operations.operation import Operation

IOperation = TypeVar('IOperation', bound=Operation)


class OperationsCacheDB(BaseCacheDB):
    """
    Database for `OperationsCache` class.
    Includes low-level idea of caching pipeline nodes using relational database.

    :param cache_folder: path to the place where cache files should be stored.
    """

    def __init__(self, cache_folder: Optional[str] = None):
        super().__init__('operations', cache_folder, False, ['pipelines_hit', 'pipelines_total', 'nodes_hit', 'nodes_total'])
        self._init_db()

    @staticmethod
    def _create_temp_for_ordered_select(cur: sqlite3.Cursor, uids: List[str]) -> str:
        """
        Creates temp table to keep order of uids while doing select operation in operations getter.

        :param cur: cursor with already installed DB connection
        :param uids: list of operations uids for keeping the order

        :return tmp_name: str name of newly created temp table
        """
        _, *other = uids
        tmp_name = f'tmp_{getpid()}'
        query = f'CREATE TEMP TABLE {tmp_name} AS SELECT 1 as id1, ? as id2'
        for num, _ in enumerate(other, 2):
            query += f' union SELECT {num} as id1, ? as id2'
        cur.execute(query, uids)
        return tmp_name

    def get_operations(self, uids: List[str]) -> List[Optional['IOperation']]:
        """
        Maps given uids to scores from DB and puts None if is not present.

        :param uids: list of operations uids to be mapped

        :return retrieved: list of operations taken from DB table with None where it wasn't present
        """
        with closing(sqlite3.connect(self.db_path)) as conn:
            with conn:
                cur = conn.cursor()
                tmp_name = self._create_temp_for_ordered_select(cur, uids)
                cur.execute((
                    f'SELECT operation FROM {tmp_name} '
                    f'LEFT JOIN {self._main_table} ON {self._main_table}.id = {tmp_name}.id2 '
                    f'ORDER BY {tmp_name}.id1;'
                ))
                retrieved = cur.fetchall()
                if self.use_stats:
                    non_null = [x for (x,) in retrieved if x is not None]
                    self._inc_eff(cur, 'nodes_hit', len(non_null))
                    if len(non_null) == len(uids):
                        self._inc_eff(cur, 'pipelines_hit')
                    self._inc_eff(cur, 'nodes_total', len(uids))
                    self._inc_eff(cur, 'pipelines_total')
                retrieved = [pickle.loads(x) if x is not None else None for (x,) in retrieved]
                return retrieved

    def add_operations(self, uid_val_lst: List[Tuple[str, 'IOperation']]):
        """
        Adds operation score to DB table via its uid

        :param uid_val_lst: list of pairs (uid -> operation) to be saved
        """
        with closing(sqlite3.connect(self.db_path)) as conn:
            with conn:
                cur = conn.cursor()
                pickled = [
                    (uid, sqlite3.Binary(pickle.dumps(val, pickle.HIGHEST_PROTOCOL)))
                    for uid, val in uid_val_lst
                ]
                cur.executemany(f'INSERT OR IGNORE INTO {self._main_table} VALUES (?, ?);', pickled)

    def _init_db(self):
        """
        Initializes DB working table.
        """
        with closing(sqlite3.connect(self.db_path)) as conn:
            with conn:
                cur = conn.cursor()
                cur.execute((
                    f'CREATE TABLE IF NOT EXISTS {self._main_table} ('
                    'id TEXT PRIMARY KEY,'
                    'operation BLOB'
                    ');'
                ))



from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline


def regression_three_depth_manual_pipeline():
    """
    Returns pipeline with the following structure:

    .. image:: img_regression_pipelines/three_depth_manual.png
      :width: 55%

    Where rf - random forest, dtreg - tree regression, knn - K nearest neighbors regression,
   """
    rfr_primary = PrimaryNode('rfr')
    knn_primary = PrimaryNode('knnreg')

    dtreg_secondary = SecondaryNode('dtreg', nodes_from=[rfr_primary])
    rfr_secondary = SecondaryNode('rfr', nodes_from=[knn_primary])

    knnreg_root = SecondaryNode('knnreg', nodes_from=[dtreg_secondary, rfr_secondary])

    pipeline = Pipeline(knnreg_root)

    return pipeline


def regression_ransac_pipeline():
    """
    Returns pipeline with the following structure:

    .. image:: img_regression_pipelines/ransac.png
      :width: 55%

    Where ransac_lin_reg - ransac algorithm
   """
    node_scaling = PrimaryNode('scaling')
    node_ransac = SecondaryNode('ransac_lin_reg', nodes_from=[node_scaling])
    node_ridge = SecondaryNode('lasso', nodes_from=[node_ransac])
    pipeline = Pipeline(node_ridge)
    return pipeline



import random

import numpy as np

from cases.credit_scoring.credit_scoring_problem import get_scoring_data, calculate_validation_metric
from fedot.core.data.data import InputData
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.pipelines.tuning.unified import PipelineTuner
from fedot.core.repository.quality_metrics_repository import ClassificationMetricsEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum

random.seed(1)
np.random.seed(1)


def get_refinement_pipeline():
    """ Create 3-level pipeline with class_decompose node """
    node_scaling = PrimaryNode('scaling')
    node_logit = SecondaryNode('logit', nodes_from=[node_scaling])
    node_decompose = SecondaryNode('class_decompose', nodes_from=[node_logit, node_scaling])
    node_rfr = SecondaryNode('rfr', nodes_from=[node_decompose])
    node_rf = SecondaryNode('rf', nodes_from=[node_rfr, node_logit])

    pipeline = Pipeline(node_rf)
    return pipeline


def get_non_refinement_pipeline():
    """ Create 3-level pipeline without class_decompose node """
    node_scaling = PrimaryNode('scaling')
    node_rf = SecondaryNode('rf', nodes_from=[node_scaling])
    node_logit = SecondaryNode('logit', nodes_from=[node_scaling])
    node_rf = SecondaryNode('rf', nodes_from=[node_logit, node_rf])
    pipeline = Pipeline(node_rf)
    return pipeline


def display_roc_auc(pipeline_to_validate, test_dataset, pipeline_name: str):
    roc_auc_metric = calculate_validation_metric(pipeline_to_validate, test_dataset)
    print(f'{pipeline_name} ROC AUC: {roc_auc_metric:.4f}')


def run_refinement_scoring_example(train_path, test_path, with_tuning=False):
    """ Function launch example with error modeling for classification task

    :param train_path: path to the csv file with training sample
    :param test_path: path to the csv file with test sample
    :param with_tuning: is it need to tune pipelines or not
    """

    task = Task(TaskTypesEnum.classification)
    train_dataset = InputData.from_csv(train_path, task=task)
    test_dataset = InputData.from_csv(test_path, task=task)

    # Get and fit pipelines
    no_decompose_c = get_non_refinement_pipeline()
    decompose_c = get_refinement_pipeline()

    no_decompose_c.fit(train_dataset)
    decompose_c.fit(train_dataset)

    # Check metrics for both pipelines
    display_roc_auc(no_decompose_c, test_dataset, 'Non decomposition pipeline')
    display_roc_auc(decompose_c, test_dataset, 'With decomposition pipeline')

    if with_tuning:
        tuner = TunerBuilder(task) \
            .with_tuner(PipelineTuner)\
            .with_metric(ClassificationMetricsEnum.ROCAUC)\
            .with_iterations(30) \
            .build(train_dataset)
        no_decompose_c = tuner.tune(no_decompose_c)
        decompose_c = tuner.tune(decompose_c)

        no_decompose_c.fit(test_dataset)
        decompose_c.fit(test_dataset)

        display_roc_auc(no_decompose_c, test_dataset, 'Non decomposition pipeline after tuning')
        display_roc_auc(decompose_c, test_dataset, 'With decomposition pipeline after tuning')


if __name__ == '__main__':
    full_path_train, full_path_test = get_scoring_data()
    run_refinement_scoring_example(full_path_train, full_path_test, with_tuning=True)



from typing import List

from fedot.core.optimisers.advisor import DefaultChangeAdvisor, RemoveType
from fedot.core.optimisers.graph import OptNode
from fedot.core.repository.operation_types_repository import get_operations_for_task


class PipelineChangeAdvisor(DefaultChangeAdvisor):
    def __init__(self, task=None):
        self.models: List[str] = get_operations_for_task(task, mode='model')
        self.data_operations: List[str] = get_operations_for_task(task, mode='data_operation')
        super().__init__(task)

    def can_be_removed(self, node: OptNode) -> RemoveType:
        operation_id = node.content['name']
        if 'exog_ts' in operation_id:
            return RemoveType.forbidden
        if 'custom' in operation_id or 'lagged' in operation_id:
            return RemoveType.with_parents
        if 'data_source' in operation_id:
            return RemoveType.with_direct_children
        return RemoveType.node_only

    def propose_change(self, node: OptNode, possible_operations: List[str]) -> List[str]:
        """
        Proposes promising candidates for node replacement
        :param node: node to propose changes for
        :param possible_operations: list of candidates for replace
        :return: list of candidates with str operations
        """
        operation_id = node.content['name']
        # data source, exog_ts and custom models replacement is useless
        if check_for_specific_operations(operation_id):
            return []

        is_model = operation_id in self.models
        similar_operations = self.models if is_model else self.data_operations

        candidates = set.intersection(set(similar_operations), set(possible_operations))

        if 'lagged' in operation_id:
            # lagged transform can be replaced only to lagged
            candidates = set.intersection({'lagged', 'sparse_lagged'}, set(possible_operations))

        if operation_id in candidates:
            # the change to the same node is not meaningful
            candidates.remove(operation_id)
        return candidates

    def propose_parent(self, node: OptNode, possible_operations: List[str]) -> List[str]:
        """
        Proposes promising candidates for new parents
        :param node: node to propose changes for
        :param possible_operations: list of candidates for replace
        :return: list of candidates with str operations
        """

        operation_id = node.content['name']
        if check_for_specific_operations(operation_id):
            # data source, exog_ts and custom models moving is useless
            return []

        parent_operations = [str(n.content['name']) for n in node.nodes_from]
        candidates = set.intersection(set(self.data_operations), set(possible_operations))

        if operation_id in candidates:
            candidates.remove(operation_id)
        if parent_operations:
            for parent_operation_id in parent_operations:
                if check_for_specific_operations(parent_operation_id):
                    # data source, exog_ts and custom models moving is useless
                    return []
                if parent_operation_id in candidates:
                    # the sequence of the same parent and child is not meaningful
                    candidates.remove(parent_operation_id)
        return candidates


def check_for_specific_operations(operation_id: str):
    if ('data_source' in operation_id or
            'exog_ts' == operation_id or 'custom' in operation_id):
        return True
    return False
